import numpy as np
import pandas as pd
#from typing import List, Series, DataFrame

def expand_label_for_meta_labeling(
    close_index: pd.Index,
    timestamp: pd.Series,
    molecule: pd.Index
) -> pd.Series:
    """
    Expand labels for meta-labeling.

    This function expands labels to incorporate meta-labeling by taking
    an event Index, a Series with the return and label of each period,
    and an Index specifying the molecules to apply the function to. It then returns a Series with the count
    of events spanning a bar for each molecule.

    :param event_index: Index of events.
    :param return_label_dataframe: Series containing returns and labels of each period.
    :param molecule_index: Index specifying molecules to apply the function on.
    :return: Series with the count of events spanning a bar for each molecule.
    """
    timestamp = timestamp.fillna(close_index[-1])
    timestamp = timestamp[timestamp >= molecule[0]]
    timestamp = timestamp.loc[:timestamp[molecule].max()]
    iloc = close_index.searchsorted(np.array([timestamp.index[0], timestamp.max()]))
    count = pd.Series(0, index=close_index[iloc[0]:iloc[1] + 1])

    for t_in, t_out in timestamp.items():
        count.loc[t_in:t_out] += 1

    return count.loc[molecule[0]:timestamp[molecule].max()]

def calculate_sample_weight(
    timestamp: pd.DataFrame,
    concurrency_events: pd.DataFrame,
    molecule: pd.Index
) -> pd.Series:
    """
    Calculate sample weight using triple barrier method.

    :param timestamp: DataFrame of events start and end for labelling.
    :param concurrency_events: Data frame of concurrent events for each event.
    :param molecule: Index that function must apply on it.
    :return: Series of sample weights.
    """
    weight = pd.Series(index=molecule)

    for t_in, t_out in timestamp.loc[weight.index].items():
        weight.loc[t_in] = (1. / concurrency_events.loc[t_in:t_out]).mean()

    return weight

def create_index_matgrix(
    bar_index: pd.Index,
    timestamp: pd.DataFrame
) -> pd.DataFrame:
    """
    Create an indicator matrix.

    :param bar_index: Index of all data.
    :param timestamp: DataFrame with starting and ending times of events.
    :return: Indicator matrix.
    """
    ind_matrix = pd.DataFrame(0, index=bar_index, columns=range(timestamp.shape[0]))

    for row in timestamp.itertuples():
        t0 = int(row.date)
        t1 = int(row.timestamp)
        ind_matrix.loc[t0:t1, row.Index] = 1

    return ind_matrix

def calculate_average_uniqueness(index_matrix: pd.DataFrame) -> pd.Series:
    """
    Calculate average uniqueness from indicator matrix.

    :param index_matrix: Indicator matrix.
    :return: Series of average uniqueness values.
    """
    concurrency = index_matrix.sum(axis=1)
    uniqueness = index_matrix.div(concurrency, axis=0)
    average_uniqueness = uniqueness[uniqueness > 0].mean()

    return average_uniqueness

def perform_sequential_bootstrap(
    index_matrix: pd.DataFrame,
    sample_length: int
) -> list:
    """
    Perform sequential bootstrap to generate a sample.

    :param index_matrix: Matrix of indicators for events.
    :param sample_length: Number of samples.
    :return: List of indices representing the sample.
    """
    if sample_length is None:
        sample_length = index_matrix.shape[1]

    phi = []

    while len(phi) < sample_length:
        average_uniqueness = pd.Series(dtype=np.float64)

        for i in index_matrix:
            index_matrix_ = index_matrix[phi + [i]]
            average_uniqueness.loc[i] = calculate_average_uniqueness(index_matrix_).iloc[-1]

        prob = average_uniqueness / average_uniqueness.sum()
        phi += [np.random.choice(index_matrix.columns, p=prob)]

    return phi

def calculate_sample_weight_absolute_return(
    timestamp: pd.DataFrame,
    concurrency_events: pd.DataFrame,
    returns: pd.DataFrame,
    molecule: pd.Index
) -> pd.Series:
    """
    Calculate sample weight using absolute returns.

    :param timestamp: DataFrame for events.
    :param concurrency_events: DataFrame that contains number of concurrent events for each event.
    :param returns: DataFrame that contains returns.
    :param molecule: Index for the calculation.
    :return: Series of sample weights.
    """
    return_ = np.log(returns).diff()
    weight = pd.Series(index=molecule)

    for t_in, t_out in timestamp.loc[weight.index].iteritems():
        weight.loc[t_in] = (return_.loc[t_in:t_out] / concurrency_events.loc[t_in:t_out]).sum()

    return weight.abs()

def sample_weight_absolute_return_meta_labeling(
    timestamp: pd.Series,
    price: pd.Series,
    molecule: pd.Index
) -> pd.Series:
    """
    Calculate sample weights using absolute returns.

    :param event_timestamps: Series containing event timestamps.
    :param price_series: Series containing prices.
    :param molecule_index: Index for the calculation.
    :return: Series of sample weights.
    """
    concurrency_events = expand_label_for_meta_labeling(price.index, timestamp, molecule)

    return_ = np.log(price).diff()
    weight = pd.Series(index=molecule, dtype=float)

    for t_in, t_out in timestamp.loc[weight.index].items():
        weight.loc[t_in] = (return_.loc[t_in:t_out] / concurrency_events.loc[t_in:t_out]).sum()

    weight = weight.abs()

    return weight * len(weight) / weight.sum()

def calculate_time_decay(
    weight: pd.Series,
    clf_last_weight: float = 1.0
) -> pd.Series:
    """
    Calculate time decay on weight.

    :param weight: Weight computed for each event.
    :param clf_last_weight: Weight of oldest observation.
    :return: Series of weights after applying time decay.
    """
    clf_weight = weight.sort_index().cumsum()

    if clf_last_weight >= 0:
        slope = (1. - clf_last_weight) / clf_weight.iloc[-1]
    else:
        slope = 1. / ((clf_last_weight + 1) * clf_weight.iloc[-1])

    const = 1. - slope * clf_weight.iloc[-1]
    clf_weight = const + slope * clf_weight
    clf_weight[clf_weight < 0] = 0

    return clf_weight



import torch
import torch.nn.functional as F
from torch import nn, Tensor
import numpy as np
from torch.nn import Module, Linear, BatchNorm1d, Tanh
from numba import cuda
from torch import autograd
import torch.nn as nn
import torch.nn.functional as F
import math
"""
In this section we have used code available online in : https://github.com/frankhan91/DeepBSDE

"""

class Equation:
    """
    Base class for defining PDE related function.

    Args:
    eqn_config (dict): dictionary containing PDE configuration parameters

    Attributes:
    dim (int): dimensionality of the problem
    total_time (float): total time horizon
    num_time_interval (int): number of time steps
    delta_t (float): time step size
    sqrt_delta_t (float): square root of time step size
    y_init (None): initial value of the function
    """

    def __init__(self, eqn_config: dict):
        self.dim = eqn_config['dim']
        self.total_time = eqn_config['total_time']
        self.num_time_interval = eqn_config['num_time_interval']
        self.delta_t = self.total_time / self.num_time_interval
        self.sqrt_delta_t = np.sqrt(self.delta_t)
        self.y_init = None

    def sample(self, num_sample: int) -> Tensor:
        """
        Sample forward SDE.

        Args:
        num_sample (int): number of samples to generate

        Returns:
        Tensor: tensor of size [num_sample, dim+1] containing samples
        """
        raise NotImplementedError

    def r_u(self, t: float, x: Tensor, y: Tensor, z: Tensor) -> Tensor:
        """
        Interest rate in the PDE.

        Args:
        t (float): current time
        x (Tensor): tensor of size [batch_size, dim] containing space coordinates
        y (Tensor): tensor of size [batch_size, 1] containing function values
        z (Tensor): tensor of size [batch_size, dim] containing gradients

        Returns:
        Tensor: tensor of size [batch_size, 1] containing generator values
        """
        raise NotImplementedError

    def h_z(self, t,x,y,z: Tensor) -> Tensor:
        """
        Function to compute H(z) in the PDE.

        Args:
        h (float): value of H function
        z (Tensor): tensor of size [batch_size, dim] containing gradients

        Returns:
        Tensor: tensor of size [batch_size, dim] containing H(z)
        """
        raise NotImplementedError

    def terminal(self, t: float, x: Tensor) -> Tensor:
        """
        Terminal condition of the PDE.

        Args:
        t (float): current time
        x (Tensor): tensor of size [batch_size, dim] containing space coordinates

        Returns:
        Tensor: tensor of size [batch_size, 1] containing terminal values
        """
        raise NotImplementedError


class PricingDefaultRisk(Equation):
  """
  Args:
  eqn_config (dict): dictionary containing PDE configuration parameters
  """
  def __init__(self, eqn_config):
    super(PricingDefaultRisk, self).__init__(eqn_config)
    self.x_init = np.ones(self.dim) * 100.0
    self.sigma = 0.2
    self.rate = 0.02   # interest rate R
    self.delta = 2.0 / 3
    self.gammah = 0.2
    self.gammal = 0.02
    self.mu_bar = 0.02
    self.K = 100
    self.vh = 50.0
    self.vl = 70.0
    self.slope = (self.gammah - self.gammal) / (self.vh - self.vl)

  def sample(self, num_sample)-> tuple:
    """
    Sample forward SDE.

    Args:
    num_sample (int): number of samples to generate

    Returns:
    tuple: tuple of two tensors: dw_sample of size [num_sample, dim, num_time_interval] and
    x_sample of size [num_sample, dim, num_time_interval+1]
    """

    dw_sample = np.random.normal(size=[num_sample, self.dim, self.num_time_interval]) * self.sqrt_delta_t
    x_sample = np.zeros([num_sample, self.dim, self.num_time_interval + 1])
    x_sample[:, :, 0] = np.ones([num_sample, self.dim]) * self.x_init
    for i in range(self.num_time_interval):
        x_sample[:, :, i + 1] = (1 + self.mu_bar * self.delta_t) * x_sample[:, :, i] + (
            self.sigma_matrix(x_sample[:, :, i]) * dw_sample[:, :, i])
    return dw_sample, x_sample

  def r_u(self, t, x, y, z)-> torch.Tensor:
    """
    Interest rate in the PDE.

    Args:
    t (float): current time
    x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
    y (torch.Tensor): tensor of size [batch_size, 1] containing function values
    z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

    Returns:
    torch.Tensor: tensor of size [batch_size, 1] containing generator values
    """
    piecewise_linear = nn.ReLU()(
        nn.ReLU()(y - self.vh) * self.slope + self.gammah - self.gammal) + self.gammal
    return ((1 - self.delta) * piecewise_linear  + self.rate)

  def h_z(self,t,x,y,z)-> torch.Tensor:
      """
      Function to compute $h^T Z$ in the PDE.

      Args:
      t (float): current time
      x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
      y (torch.Tensor): tensor of size [batch_size, 1] containing function value
      z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

      Returns:
      torch.Tensor: tensor of size [batch_size, 1] containing H(z)
      """
      return torch.zeros((x.size()[0],1))

  def sigma_matrix(self,x):
    return self.sigma * x


  def terminal(self, t, x)-> torch.Tensor:
    """
    Terminal condition of the PDE.

    Args:
    t (float): current time
    x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates

    Returns:
    torch.Tensor: tensor of size [batch_size, 1] containing terminal values
    """
    return nn.ReLU()(torch.min(x, 1)[0])
  def terminal_for_sample(self, x)-> torch.Tensor:
    """
    Terminal condition of the PDE.

    Args:
    x (torch.Tensor): tensor of size [num_sample,batch_size, dim] containing space coordinates

    Returns:
    torch.Tensor: tensor of size [num_sample ,batch_size, 1] containing terminal values
    """
    return nn.ReLU()(torch.min(x, 2 , keepdim= True)[0])


class HJBLQ(Equation):
  """

  Args:
  eqn_config (dict): dictionary containing PDE configuration parameters
  """

  def __init__(self, eqn_config: dict):
    super().__init__(eqn_config)

    self.x_init = np.zeros(self.dim)
    self.sigma = np.sqrt(2.0)
    self.lambd = 1.0

  def sample(self, num_sample: int) -> tuple:
    """
    Sample forward SDE.

    Args:
    num_sample (int): number of samples to generate

    Returns:
    tuple: tuple of two tensors: dw_sample of size [num_sample, dim, num_time_interval] and
    x_sample of size [num_sample, dim, num_time_interval+1]
    """
    dw_sample = np.random.normal(size=[num_sample, self.dim, self.num_time_interval]) * self.sqrt_delta_t
    x_sample = np.zeros([num_sample, self.dim, self.num_time_interval + 1])
    x_sample[:, :, 0] = np.ones([num_sample, self.dim]) * self.x_init
    for i in range(self.num_time_interval):
        x_sample[:, :, i + 1] = x_sample[:, :, i] + self.sigma * dw_sample[:, :, i]
    return dw_sample, x_sample

  def r_u(self, t: float, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
    """
    Interest rate in the PDE.

    Args:
    t (float): current time
    x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
    y (torch.Tensor): tensor of size [batch_size, 1] containing function values
    z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

    Returns:
    torch.Tensor: tensor of size [batch_size, 1] containing generator values
    """
    return torch.ones((x.size()[0],)) * 0

  def h_z(self, t: float, x: torch.Tensor,y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
    """
    Function to compute <h,z> in the PDE.

    Args:
    t (float): current time
    x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
    y (torch.Tensor): tensor of size [batch_size, 1] containing function value
    z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

    Returns:
    torch.Tensor: tensor of size [batch_size, 1] containing H(z)
    """
    return torch.sum(torch.square(z), dim=1)/ (self.sigma**2)

  def terminal(self, t: float, x: torch.Tensor) -> torch.Tensor:
    """
    Terminal condition of the PDE.

    Args:
    t (float): current time
    x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates

    Returns:
    torch.Tensor: tensor of size [batch_size, 1] containing terminal values
    """
    return torch.log(0.5 * (1 + torch.norm(x, dim=1)**2))
  def sigma_matrix(self,x):
    return self.sigma
def terminal_for_sample(self, x)-> torch.Tensor:
    """
    Terminal condition of the PDE.

    Args:
    x (torch.Tensor): tensor of size [num_sample,batch_size, dim] containing space coordinates

    Returns:
    torch.Tensor: tensor of size [num_sample ,batch_size, 1] containing terminal values
    """
    return torch.sum(x ** 2, dim=2 ,keepdim=True )

class BlackScholesBarenblatt(Equation):
  """

  Args:
  eqn_config (dict): dictionary containing PDE configuration parameters
  """

  def __init__(self, eqn_config: dict):
    super().__init__(eqn_config)
    self.x_init = np.ones(self.dim) * np.array([1.0 / (1.0 + i % 2) for i in range(self.dim)])
    self.sigma = 0.4
    self.rate = 0.05   # interest rate R
    self.mu_bar = 0.0

  def sample(self, num_sample: int) -> tuple:
    """
    Sample forward SDE.

    Args:
    num_sample (int): number of samples to generate

    Returns:
    tuple: tuple of two tensors: dw_sample of size [num_sample, dim, num_time_interval] and
    x_sample of size [num_sample, dim, num_time_interval+1]
    """
    dw_sample = np.random.normal(size=(num_sample, self.dim, self.num_time_interval)) * self.sqrt_delta_t
    x_sample = np.zeros((num_sample, self.dim, self.num_time_interval + 1))
    x_sample[:, :, 0] = np.ones((num_sample, self.dim)) * self.x_init
    for i in range(self.num_time_interval):
        x_sample[:, :, i + 1] = (1 + self.mu_bar * self.delta_t) * x_sample[:, :, i] + (
            self.sigma_matrix(x_sample[:, :, i]) * dw_sample[:, :, i])
    return dw_sample, x_sample

  def r_u(self, t: float, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
    """
    Interest rate in the PDE.

    Args:
    t (float): current time
    x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
    y (torch.Tensor): tensor of size [batch_size, 1] containing function values
    z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

    Returns:
    torch.Tensor: tensor of size [batch_size, 1] containing generator values
    """
    return torch.ones((x.size()[0],)) * self.rate

  def h_z(self, t: float, x: torch.Tensor,y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
    """
    Function to compute <h,z> in the PDE.

    Args:
    t (float): current time
    x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
    y (torch.Tensor): tensor of size [batch_size, 1] containing function value
    z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

    Returns:
    torch.Tensor: tensor of size [batch_size, 1] containing H(z)
    """
    return -1 * torch.sum(z, dim=1) * self.rate / self.sigma

  def terminal(self, t: float, x: torch.Tensor) -> torch.Tensor:
    """
    Terminal condition of the PDE.

    Args:
    t (float): current time
    x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates

    Returns:
    torch.Tensor: tensor of size [batch_size, 1] containing terminal values
    """
    return torch.sum(x ** 2, dim=1)
  def sigma_matrix(self,x):
    return self.sigma * x
  def terminal_for_sample(self, x)-> torch.Tensor:
    """
    Terminal condition of the PDE.

    Args:
    x (torch.Tensor): tensor of size [num_sample,batch_size, dim] containing space coordinates

    Returns:
    torch.Tensor: tensor of size [num_sample ,batch_size, 1] containing terminal values
    """
    return torch.sum(x ** 2, dim=2 ,keepdim=True )

class PricingDiffRate(Equation):
    """
    Nonlinear Black-Scholes equation with different interest rates for borrowing and lending
    in Section 4.4 of Comm. Math. Stat. paper doi.org/10.1007/s40304-017-0117-6
    """
    def __init__(self, eqn_config):
        super(PricingDiffRate, self).__init__(eqn_config)
        self.x_init = np.ones(self.dim) * 100
        self.sigma = 0.2
        self.mu_bar = 0.06
        self.rl = 0.04
        self.rb = 0.06
        self.alpha = 1.0 / self.dim

    def sample(self, num_sample):
        dw_sample = np.random.normal(size=[num_sample, self.dim, self.num_time_interval]) * self.sqrt_delta_t
        x_sample = np.zeros([num_sample, self.dim, self.num_time_interval + 1])
        x_sample[:, :, 0] = np.ones([num_sample, self.dim]) * self.x_init
        factor = np.exp((self.mu_bar-(self.sigma**2)/2)*self.delta_t)
        for i in range(self.num_time_interval):
            x_sample[:, :, i + 1] = (factor * np.exp(self.sigma * dw_sample[:, :, i])) * x_sample[:, :, i]
        return dw_sample, x_sample

    def r_u(self, t: float, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
      temp = torch.sum(z, dim= 1) / self.sigma - y
      return torch.where(temp > 0 , self.rb , self.rl)
    def h_z(self, t: float, x: torch.Tensor,y: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
      """
      Function to compute <h,z> in the PDE.

      Args:
      t (float): current time
      x (torch.Tensor): tensor of size [batch_size, dim] containing space coordinates
      y (torch.Tensor): tensor of size [batch_size, 1] containing function value
      z (torch.Tensor): tensor of size [batch_size, dim] containing gradients

      Returns:
      torch.Tensor: tensor of size [batch_size, 1] containing H(z)
      """
      temp = torch.sum(z, dim= 1) / self.sigma - y
      return torch.where(temp > 0 , (self.mu_bar - self.rb) * torch.sum(z, dim=1) / self.sigma ,(self.mu_bar - self.rl) * torch.sum(z, dim=1) / self.sigma)


    def terminal(self, t, x):
        temp = torch.max(x, 1)[0]
        return nn.ReLU()(temp - 120) - 2*nn.ReLU()(temp - 150)
    def sigma_matrix(self,x):
      return self.sigma * x
    def terminal_for_sample(self, x)-> torch.Tensor:
        """
        Terminal condition of the PDE.

        Args:
        x (torch.Tensor): tensor of size [num_sample,batch_size, dim] containing space coordinates

        Returns:
        torch.Tensor: tensor of size [num_sample ,batch_size, 1] containing terminal values
        """
        temp = torch.max(x, 2 , keep_dim = True)[0]
        return torch.maximum(temp - 120, torch.tensor(0))



import numpy as np
import pandas as pd

def bet_timing(target_positions: pd.Series) -> pd.Index:
    """
    Determine the timing of bets when positions flatten or flip.

    :param target_positions: Series of target positions.
    :return: Index of bet timing.
    """
    zero_positions = target_positions[target_positions == 0].index

    lagged_non_zero_positions = target_positions.shift(1)
    lagged_non_zero_positions = lagged_non_zero_positions[lagged_non_zero_positions != 0].index

    bets = zero_positions.intersection(lagged_non_zero_positions)
    zero_positions = target_positions.iloc[1:] * target_positions.iloc[:-1].values
    bets = bets.union(zero_positions[zero_positions < 0].index).sort_values()

    if target_positions.index[-1] not in bets:
        bets = bets.append(target_positions.index[-1:])

    return bets

def calculate_holding_period(target_positions: pd.Series) -> tuple:
    """
    Derive average holding period (in days) using the average entry time pairing algorithm.

    :param target_positions: Series of target positions.
    :return: Tuple containing holding period DataFrame and mean holding period.
    """
    hold_period, time_entry = pd.DataFrame(columns=['dT', 'w']), 0.0
    position_difference = target_positions.diff()
    time_difference = (target_positions.index - target_positions.index[0]) / np.timedelta64(1, 'D')

    for i in range(1, target_positions.shape[0]):
        if position_difference.iloc[i] * target_positions.iloc[i - 1] >= 0:
            if target_positions.iloc[i] != 0:
                time_entry = (time_entry * target_positions.iloc[i - 1] + time_difference[i] * position_difference.iloc[i]) / target_positions.iloc[i]
        else:
            if target_positions.iloc[i] * target_positions.iloc[i - 1] < 0:
                hold_period.loc[target_positions.index[i], ['dT', 'w']] = (time_difference[i] - time_entry, abs(target_positions.iloc[i - 1]))
                time_entry = time_difference[i]
            else:
                hold_period.loc[target_positions.index[i], ['dT', 'w']] = (time_difference[i] - time_entry, abs(position_difference.iloc[i]))

    if hold_period['w'].sum() > 0:
        mean_holding_period = (hold_period['dT'] * hold_period['w']).sum() / hold_period['w'].sum()
    else:
        mean_holding_period = np.nan

    return hold_period, mean_holding_period

def calculate_hhi_concentration(returns: pd.Series) -> tuple:
    """
    Calculate the HHI concentration measures.

    :param returns: Series of returns.
    :return: Tuple containing positive returns HHI, negative returns HHI, and time-concentrated HHI.
    """
    returns_hhi_positive = calculate_hhi(returns[returns >= 0])
    returns_hhi_negative = calculate_hhi(returns[returns < 0])
    time_concentrated_hhi = calculate_hhi(returns.groupby(pd.Grouper(freq='M')).count())

    return returns_hhi_positive, returns_hhi_negative, time_concentrated_hhi

def calculate_hhi(bet_returns: pd.Series) -> float:
    """
    Calculate the Herfindahl-Hirschman Index (HHI) concentration measure.

    :param bet_returns: Series of bet returns.
    :return: Calculated HHI value.
    """
    if bet_returns.shape[0] <= 2:
        return np.nan

    weight = bet_returns / bet_returns.sum()
    hhi_ = (weight ** 2).sum()
    hhi_ = (hhi_ - bet_returns.shape[0] ** -1) / (1.0 - bet_returns.shape[0] ** -1)

    return hhi_

def compute_drawdowns_time_under_water(series: pd.Series, dollars: bool = False) -> tuple:
    """
    Compute series of drawdowns and the time under water associated with them.

    :param series: Series of returns or dollar performance.
    :param dollars: Whether the input series represents returns or dollar performance.
    :return: Tuple containing drawdown series, time under water series, and drawdown analysis DataFrame.
    """
    series_df = series.to_frame('PnL').reset_index(names='Datetime')
    series_df['HWM'] = series.expanding().max().values

    def process_groups(group):
        if len(group) <= 1:
            return None

        result = pd.Series()
        result.loc['Start'] = group['Datetime'].iloc[0]
        result.loc['Stop'] = group['Datetime'].iloc[-1]
        result.loc['HWM'] = group['HWM'].iloc[0]
        result.loc['Min'] = group['PnL'].min()
        result.loc['Min. Time'] = group['Datetime'][group['PnL'] == group['PnL'].min()].iloc[0]

        return result

    groups = series_df.groupby('HWM')
    drawdown_analysis = pd.DataFrame()

    for _, group in groups:
        drawdown_analysis = drawdown_analysis.append(process_groups(group), ignore_index=True)

    if dollars:
        drawdown = drawdown_analysis['HWM'] - drawdown_analysis['Min']
    else:
        drawdown = 1 - drawdown_analysis['Min'] / drawdown_analysis['HWM']

    drawdown.index = drawdown_analysis['Start']
    drawdown.index.name = 'Datetime'

    time_under_water = ((drawdown_analysis['Stop'] - drawdown_analysis['Start']) / np.timedelta64(1, 'Y')).values
    time_under_water = pd.Series(time_under_water, index=drawdown_analysis['Start'])

    return drawdown, time_under_water, drawdown_analysis



from copy import deepcopy
from typing import Generator, Optional, Tuple, Union, Dict, List, Any, Set
from joblib import Parallel, delayed
import pandas as pd
import numpy as np
import warnings
from sklearn.exceptions import ConvergenceWarning

from .cross_validator_interface import CrossValidator

class PurgedKFold(CrossValidator):

    @staticmethod
    def filtered_training_indices_with_embargo(
        data_info_range: pd.Series,
        test_time_range: pd.Series,
        embargo_fraction: float = 0,
        continous_test_times: bool = False,
    ) -> pd.Series:
        """
        Purge observations in the training set with embargo.

        Finds the training set indices based on the information on each record
        and the test set range. It purges the training set of observations that
        overlap with the test set in the time dimension and adds an embargo period
        to further prevent potential information leakage.

        .. math::
            \text{embargo\_length} = \text{len(data\_info\_range)} \times \text{embargo\_fraction}

        :param data_info_range: Series detailing the information range for each record.
            - *data_info_range.index*: Time when the information extraction started.
            - *data_info_range.value*: Time when the information extraction ended.
        :type data_info_range: pd.Series
        :param test_time_range: Series containing times for the test dataset.
        :type test_time_range: pd.Series
        :param embargo_fraction: Fraction of the dataset trailing the test observations to exclude from training.
        :type embargo_fraction: float
        :param continuous_test_times: If set to True, considers the test time range as continuous.
        :type continuous_test_times: bool

        :return: Series of filtered training data after applying embargo.
        :rtype: pd.Series
        """

        indices_to_drop: Set[int] = set()
        embargo_length = int(len(data_info_range) * embargo_fraction)
        sorted_test_time_range = test_time_range.sort_index().copy()

        if not continous_test_times:
            sorted_test_time_range = pd.DataFrame({
                'start' : sorted_test_time_range.index,
                'end' : sorted_test_time_range.values
            })
            # Identify where the new range starts immediately after or before the previous range ends
            gaps = sorted_test_time_range['start'] > sorted_test_time_range['end'].shift(1)
            # Cumulative sum to identify contiguous blocks
            blocks = gaps.cumsum()
            # Aggregate to find the min start and max end for each block
            effective_test_time_range = sorted_test_time_range.groupby(blocks).agg({'start': 'min', 'end': 'max'})
            effective_test_time_range = pd.Series(effective_test_time_range['end'].values, index=effective_test_time_range['start'])

        else:
            effective_test_time_range = pd.Series(sorted_test_time_range.iloc[-1], index=[sorted_test_time_range.index[0]])

        if embargo_length == 0:
            embargoed_data_info_range = pd.Series(effective_test_time_range.values, index=effective_test_time_range.values)

        else:
            effective_sample = data_info_range.loc[effective_test_time_range.index.min():].copy().drop_duplicates()
            embargoed_data_info_range = pd.Series(effective_sample.values, index=effective_sample.values)
            embargoed_data_info_range = embargoed_data_info_range.shift(-embargo_length).fillna(embargoed_data_info_range.values[-1])   

        effective_ranges = pd.Series(embargoed_data_info_range.loc[effective_test_time_range].values, index=effective_test_time_range.index)

        for start_ix, end_ix_embargoed in effective_ranges.items():

            indices_to_drop.update(
                data_info_range[
                    ((start_ix <= data_info_range.index) & (data_info_range.index <= end_ix_embargoed)) |
                    ((start_ix <= data_info_range) & (data_info_range <= end_ix_embargoed)) |
                    ((data_info_range.index <= start_ix) & (end_ix_embargoed <= data_info_range))
                ].index
            )

        return data_info_range.drop(indices_to_drop)


    def __init__(
            self,
            n_splits: int,
            times: Union[pd.Series, Dict[str, pd.Series]],
            embargo: float = 0,
    ) -> None:
        """
        Purged k-fold cross-validation to prevent information leakage.

        Implements a cross-validation strategy where each fold is purged
        of observations overlapping with the training set in the time dimension.
        An embargo period is also introduced to further prevent potential
        information leakage.

        Attributes:
            n_splits (int): Number of splits/folds.
            times (Union[pd.Series, Dict[str, pd.Series]]): Series or dict containing time data.
            embargo (float): The embargo period.
            is_multiple_datasets (bool): True if `times` is a dict, else False.

        :param n_splits: Number of splits or folds.
        :type n_splits: int

        :param times: Series detailing the information range for each record.
            - *times.index*: Time when the information extraction started.
            - *times.value*: Time when the information extraction ended.
        :type times: pd.Series

        :param embargo: The embargo period to further prevent potential
                        information leakage.
        :type embargo: float
        """
        self.n_splits = n_splits
        self.times = times
        self.embargo = embargo
        self.is_multiple_datasets = isinstance(times, dict)


    def get_n_splits(
        self,
        data: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]] = None,
        labels: Optional[Union[pd.Series, Dict[str, pd.Series]]] = None,
        groups: Optional[np.ndarray] = None
    ) -> int:
        """
        Return number of splits.

        :param data: Dataset or dictionary of datasets.
        :type data: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]]

        :param labels: Labels or dictionary of labels.
        :type labels: Optional[Union[pd.Series, Dict[str, pd.Series]]]

        :param groups: Group labels for the samples.
        :type groups: Optional[np.ndarray]

        :return: Number of splits.
        :rtype: int
        """
        return self.n_splits


    def _validate_input(
        self,
        single_times: pd.Series,
        single_data: pd.DataFrame
    ) -> None:
        """
        Validate that the input data and times share the same index.

        This function checks if the provided data and its corresponding times
        have the same index. If they do not match, it raises a `ValueError`.

        :param single_times: Time series data to be validated.
        :type single_times: pd.Series
        :param single_data: Dataset with which the times should align.
        :type single_data: pd.DataFrame
        :raises ValueError: If the indices of the data and times do not match.
        :return: None
        """
        if not single_data.index.equals(single_times.index):
            raise ValueError('Data and through date values must have the same index')


    def _get_train_indices(
        self,
        test_indices: np.ndarray,
        single_times: pd.Series,
        continous_test_times: bool = False,
    ) -> np.ndarray:
        """
        Obtain the training indices considering purging and embargo.

        This function retrieves the training set indices based on the given test indices
        while considering the purging and embargo strategy.

        :param test_indices: Indices used for the test set.
        :type test_indices: np.ndarray
        :param single_times: Time series data used for purging and embargo.
        :type single_times: pd.Series
        :return: Training indices after applying purging and embargo.
        :rtype: np.ndarray
        """
        test_time_range = single_times.iloc[test_indices]
        train_times = self.filtered_training_indices_with_embargo(
            single_times, test_time_range, self.embargo, continous_test_times
        )
        return pd.Series(train_times.index).map(single_times.index.get_loc).tolist()


    def _single_split(
        self,
        single_times: pd.Series,
        single_data: pd.DataFrame,
    ) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
        """
        Split the data into train and test sets.

        This function splits the data for a single dataset considering purging and embargo.

        :param single_times: Time series data used for purging and embargo.
        :type single_times: pd.Series
        :param single_data: Dataset to split.
        :type single_data: pd.DataFrame
        :return: Train and test indices.
        :rtype: Generator[Tuple[np.ndarray, np.ndarray], None, None]
        """
        self._validate_input(single_times, single_data)

        indices = np.arange(len(single_data))

        for test_indices in np.array_split(indices, self.n_splits):
            train_indices = self._get_train_indices(test_indices, single_times, True)
            yield train_indices, test_indices


    def split(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        labels: Optional[Union[pd.Series, Dict[str, pd.Series]]] = None,
        groups: Optional[np.ndarray] = None
    ) -> Union[
        Generator[Tuple[np.ndarray, np.ndarray], None, None],
        Generator[Tuple[str, Tuple[np.ndarray, np.ndarray]], None, None]
    ]:
        """
        Split multiple datasets into train and test sets.

        This function either splits a single dataset or multiple datasets considering
        purging and embargo.

        :param data: Dataset or dictionary of datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
        :param labels: Labels corresponding to the datasets, if available.
        :type labels: Optional[Union[pd.Series, Dict[str, pd.Series]]]
        :param groups: Group information, if available.
        :type groups: Optional[np.ndarray]
        :return: Train and test indices or key with train and test indices for multiple datasets.
        :rtype: Union[Generator[Tuple[np.ndarray, np.ndarray], None, None],
                    Generator[Tuple[str, Tuple[np.ndarray, np.ndarray]], None, None]]
        """
        if self.is_multiple_datasets:
            for key in self.times:
                for train_indices, test_indices in self._single_split(
                    self.times[key], data[key]
                ):
                    yield key, (train_indices, test_indices)
        else:
            for train_indices, test_indices in self._single_split(
                self.times, data
            ):
                yield train_indices, test_indices


    def _single_backtest_paths(
        self,
        single_times: pd.Series,
        single_data: pd.DataFrame,
    ) -> Dict[str, List[Dict[str, np.ndarray]]]:
        """
        Generate backtest paths based on training and testing indices.

        This function first validates the input data and times. Then, it generates
        the training and testing indices for backtesting. These paths are organized
        into a dictionary with a designated name for each backtest path.

        :param single_times: Time series data for validation.
        :type single_times: pd.Series
        :param single_data: Dataset with which the times should align.
        :type single_data: pd.DataFrame
        :return: Dictionary containing the backtest paths with training and testing indices.
        :rtype: Dict[str, List[Dict[str, np.ndarray]]]
        """
        self._validate_input(single_times, single_data)

        path_data = []
        paths = {}

        for train_indices, test_indices in self._single_split(single_times, single_data):
            path_data.append({
                "Train": np.array(train_indices),
                "Test": test_indices,
            })

        paths['Path 1'] = path_data

        return paths


    def backtest_paths(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
    ) -> Union[
        Dict[str, List[Dict[str, np.ndarray]]],
        Dict[str, Dict[str, List[Dict[str, List[np.ndarray]]]]]
    ]:
        """
        Generate backtest paths for single or multiple datasets.

        This function checks whether multiple datasets are being used. If so, it iterates through each
        dataset, generating backtest paths using the `_single_backtest_paths` method. Otherwise, it directly
        returns the backtest paths for the single dataset.

        :param data: Input data on which the backtest paths are based.
                    Can be either a single DataFrame or a dictionary of DataFrames for multiple datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]

        :return: A dictionary where each key is a backtest path name, and the value is
                a list of dictionaries with train and test index arrays. For multiple datasets,
                a nested dictionary structure is returned.
        :rtype: Union[
            Dict[str, List[Dict[str, np.ndarray]]],
            Dict[str, Dict[str, List[Dict[str, List[np.ndarray]]]]]
        ]

        """
        if self.is_multiple_datasets:
            multiple_paths = {}
            for key in self.times:
                multiple_paths[key] = self._single_backtest_paths(self.times[key], data[key])

            return multiple_paths
        else:

            return self._single_backtest_paths(self.times, data)


    def _single_backtest_predictions(
        self,
        single_estimator: Any,
        single_times: pd.Series,
        single_data: pd.DataFrame,
        single_labels: pd.Series,
        single_weights: Optional[np.ndarray] = None,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Dict[str, np.ndarray]:
        """
        Obtain predictions for a single dataset during backtesting.

        This function leverages parallel computation to train and predict on different train-test splits
        of a single dataset using a given estimator. It utilizes the `_single_split` method to generate
        the train-test splits.

        :param single_estimator: Estimator or model to be trained and used for predictions.
        :type single_estimator: Any
        :param single_times: Timestamps for the single dataset.
        :type single_times: pd.Series
        :param single_data: Data of the single dataset.
        :type single_data: pd.DataFrame
        :param single_labels: Labels corresponding to the single dataset.
        :type single_labels: pd.Series
        :param single_weights: Weights for the observations in the single dataset.
                            Defaults to equally weighted if not provided.
        :type single_weights: np.ndarray, optional
        :param predict_probability: If True, predict the probability of forecasts.
        :type predict_probability: bool
        :param n_jobs: The number of jobs to run in parallel. Default is 1.
        :type n_jobs: int, optional
        :return: Predictions structured in a dictionary for the backtest paths.
        :rtype: Dict[str, np.ndarray]
        """
        self._validate_input(single_times, single_data)

        if single_weights is None:
            single_weights = np.ones(len(single_data))

        def train_test_single_estimator(
            single_estimator_: Any,
            train_indices: np.ndarray,
            test_indices: np.ndarray
        ) -> np.ndarray:
            X_train = single_data.iloc[train_indices]
            y_train = single_labels.iloc[train_indices]
            weights_train = single_weights[train_indices]

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=ConvergenceWarning)
                try:
                    single_estimator_.fit(X_train, y_train, sample_weight=weights_train)

                except TypeError:
                    single_estimator_.fit(X_train, y_train)

            X_test = single_data.iloc[test_indices]

            if predict_probability:
                return single_estimator_.predict_proba(X_test)

            else:
                return single_estimator_.predict(X_test)

        path_data = Parallel(n_jobs=n_jobs)(
            delayed(train_test_single_estimator)(
                deepcopy(single_estimator), train_indices, test_indices
            ) for train_indices, test_indices in self._single_split(single_times, single_data)
        )

        paths_predictions = {'Path 1': np.concatenate(path_data)}

        return paths_predictions


    def backtest_predictions(
        self,
        estimator: Union[Any, Dict[str, Any]],
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        labels: Union[pd.Series, Dict[str, pd.Series]],
        sample_weights: Optional[Union[np.ndarray, Dict[str, np.ndarray]]] = None,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Union[Dict[str, np.ndarray], Dict[str, Dict[str, np.ndarray]]]:
        """
        Generate backtest predictions for single or multiple datasets.

        For each dataset, this function leverages the `_single_backtest_predictions` method to obtain
        predictions for different train-test splits using the given estimator.

        :param estimator: Model or estimator to be trained and used for predictions.
                        Can be a single estimator or a dictionary of estimators for multiple datasets.
        :type estimator: Union[Any, Dict[str, Any]]
        :param data: Input data for training and testing. Can be a single dataset or
                    a dictionary of datasets for multiple datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
        :param labels: Target labels for training and testing. Can be a single series or
                    a dictionary of series for multiple datasets.
        :type labels: Union[pd.Series, Dict[str, pd.Series]]
        :param sample_weights: Weights for the observations in the dataset(s).
                            Can be a single array or a dictionary of arrays for multiple datasets.
                            Defaults to None, which means equal weights for all observations.
        :type sample_weights: Optional[Union[np.ndarray, Dict[str, np.ndarray]]]
        :param predict_probability: If True, predict the probability of forecasts.
        :type predict_probability: bool
        :param n_jobs: The number of jobs to run in parallel. Default is 1.
        :type n_jobs: int, optional
        :return: Backtest predictions structured in a dictionary (or nested dictionaries for multiple datasets).
        :rtype: Union[Dict[str, np.ndarray], Dict[str, Dict[str, np.ndarray]]]
        """

        if self.is_multiple_datasets:
            multiple_paths_predictions = {}
            for key in self.times:
                multiple_paths_predictions[key] = self._single_backtest_predictions(
                    estimator[key], self.times[key], data[key], labels[key],
                    sample_weights[key] if sample_weights else None, predict_probability, n_jobs
                )
            return multiple_paths_predictions

        return self._single_backtest_predictions(
            estimator, self.times, data, labels, sample_weights, predict_probability, n_jobs
        )



import numpy as np
import pandas as pd
from scipy import stats
from typing import List, Tuple


def calculate_t_value_linear_regression(
        price: pd.Series
) -> float:
    """
    Calculate the t-value of a linear trend.

    This function computes the t-value of a linear trend in a time series of prices. 
    The t-value is calculated as the ratio of the slope to the standard error of the regression.

    :param price: Time series of prices as a Pandas Series.
    :return: Calculated t-value as a float.
    """
    x = np.arange(price.shape[0])
    ols = stats.linregress(x, price.values)
    t_value = ols.slope / ols.stderr
    
    return t_value


def find_trend_using_trend_scanning(
        molecule: pd.Index,
        close: pd.Series,
        span: Tuple[int, int]
) -> pd.DataFrame:
    """
    Implement the trend scanning method to find trends.

    This function identifies trends in a time series of prices using the trend scanning method. 
    It calculates the t-value for linear regression of price over a range of spans and 
    identifies the span with the maximum absolute t-value as the trend. The sign of the t-value 
    indicates the direction of the trend.

    :param molecule: Index of observations to label as a Pandas Index.
    :param close: Time series of prices as a Pandas Series.
    :param span: Range of span lengths to evaluate for the maximum absolute t-value as a tuple.
    :return: DataFrame containing trend information with columns ['End Time', 't-Value', 'Trend'].
    """
    outputs = pd.DataFrame(index=molecule, columns=['End Time', 't-Value', 'Trend'])
    spans = range(*span)

    for index in molecule:
        t_values = pd.Series(dtype='float64')
        location = close.index.get_loc(index)

        if location + max(spans) > close.shape[0]:
            continue

        for span in spans:
            tail = close.index[location + span - 1]
            window_prices = close.loc[index:tail]
            t_values.loc[tail] = calculate_t_value_linear_regression(window_prices)

        tail = t_values.replace([-np.inf, np.inf, np.nan], 0).abs().idxmax()
        outputs.loc[index, ['End Time', 't-Value', 'Trend']] = t_values.index[-1], t_values[tail], np.sign(t_values[tail])

    outputs['End Time'] = pd.to_datetime(outputs['End Time'])
    outputs['Trend'] = pd.to_numeric(outputs['Trend'], downcast='signed')

    return outputs.dropna(subset=['Trend'])



import numpy as np
from typing import Tuple

def drift_volatility_burst(
    bubble_length: int,
    a_before: float,
    a_after: float,
    b_before: float,
    b_after: float,
    alpha: float,
    beta: float, 
    explosion_filter_width: float = 0.1,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the drift and volatility for a burst scenario.

    The drift and volatility are calculated based on:
    .. math::
        drift = \frac{a_{value}}{denominator^\alpha}
        volatility = \frac{b_{value}}{denominator^\beta}

    where:
    .. math::
        denominator = |step - 0.5|

    :param bubble_length: The length of the bubble.
    :param a_before: 'a' value before the mid-point.
    :param a_after: 'a' value after the mid-point.
    :param b_before: 'b' value before the mid-point.
    :param b_after: 'b' value after the mid-point.
    :param alpha: Exponent for the drift calculation.
    :param beta: Exponent for the volatility calculation.
    :param explosion_filter_width: Width of the area around the explosion that denominators won't exceed. 
    :return: A tuple containing the drift and volatility arrays.
    """
    steps = np.linspace(0, 1, bubble_length)

    # Create two boolean masks identifying the indices where the values are within the specified range
    before_mask = (steps >= (0.5 - explosion_filter_width)) & (steps < 0.5)
    after_mask = (steps > 0.5) & (steps <= (0.5 + explosion_filter_width))

    # Replace the values at these indices with 0.5 - explosion_filter_width
    steps[before_mask] = 0.5 - explosion_filter_width
    steps[after_mask] = 0.5 + explosion_filter_width

    a_values = np.where(steps <= 0.5, a_before, a_after)
    b_values = np.where(steps <= 0.5, b_before, b_after)

    denominators = np.abs(steps - 0.5)
    denominators[steps == 0.5] = np.nan  # Set the denominator to NaN for step == 0.5

    drifts = a_values / denominators ** alpha
    volatilities = b_values / denominators ** beta

    # Fill NaN values with preceding values
    nan_mask = np.isnan(denominators)
    if np.sum(nan_mask) > 0:
        drifts[nan_mask] = 0
        volatilities[nan_mask] = volatilities[np.where(nan_mask)[0][0] - 1]

    return drifts, volatilities



from collections import Counter
from math import log2
from typing import Dict, Tuple


def probability_mass_function(
    message: str,
    approximate_word_length: int
) -> Dict[str, float]:
    """
    Calculate Probability Mass Function.

    :param message: Input encoded message
    :type message: str
    :param approximate_word_length: Approximation of word length
    :type approximate_word_length: int
    :return: Probability Mass Function
    :rtype: dict
    """
    library = Counter(message[i:i + approximate_word_length]
                     for i in range(len(message) - approximate_word_length + 1))

    denominator = float(len(message) - approximate_word_length)
    probability_mass_function_ = {key: len(library[key]) / denominator
                                 for key in library}

    return probability_mass_function_




from typing import Tuple, Union

import pandas as pd

from RiskLabAI.data.structures.imbalance_bars import ExpectedImbalanceBars, FixedImbalanceBars
from RiskLabAI.data.structures.run_bars import ExpectedRunBars, FixedRunBars
from RiskLabAI.data.structures.standard_bars import StandardBars
from RiskLabAI.data.structures.time_bars import TimeBars

from RiskLabAI.utils.constants import CUMULATIVE_DOLLAR, CUMULATIVE_VOLUME, CUMULATIVE_TICKS


class BarsInitializerController:
    """
    Controller for initializing various types of bars.
    """

    def __init__(self):
        self.method_name_to_method = {
            "expected_dollar_imbalance_bars": self.initialize_expected_dollar_imbalance_bars,
            "expected_volume_imbalance_bars": self.initialize_expected_volume_imbalance_bars,
            "expected_tick_imbalance_bars": self.initialize_expected_tick_imbalance_bars,
            "fixed_dollar_imbalance_bars": self.initialize_fixed_dollar_imbalance_bars,
            "fixed_volume_imbalance_bars": self.initialize_fixed_volume_imbalance_bars,
            "fixed_tick_imbalance_bars": self.initialize_fixed_tick_imbalance_bars,
            "expected_dollar_run_bars": self.initialize_expected_dollar_run_bars,
            "expected_volume_run_bars": self.initialize_expected_volume_run_bars,
            "expected_tick_run_bars": self.initialize_expected_tick_run_bars,
            "fixed_dollar_run_bars": self.initialize_fixed_dollar_run_bars,
            "fixed_volume_run_bars": self.initialize_fixed_volume_run_bars,
            "fixed_tick_run_bars": self.initialize_fixed_tick_run_bars,
            "dollar_standard_bars": self.initialize_dollar_standard_bars,
            "volume_standard_bars": self.initialize_volume_standard_bars,
            "tick_standard_bars": self.initialize_tick_standard_bars,
            "time_bars": self.initialize_time_bars
        }

    @staticmethod
    def initialize_expected_dollar_imbalance_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            expected_ticks_number_bounds: Tuple[float] = None,
            analyze_thresholds: bool = False
    ) -> ExpectedImbalanceBars:
        """
        Initialize expected dollar imbalance bars.

        :param window_size_for_expected_n_ticks_estimation: The window size for estimating the expected number of ticks.
        :param window_size_for_expected_imbalance_estimation: The window size for estimating the expected imbalance.
        :param initial_estimate_of_expected_n_ticks_in_bar: The initial estimate for the expected number of ticks in a bar.
        :param expected_ticks_number_bounds: Bounds for the expected number of ticks in a bar.
        :param analyze_thresholds: Flag indicating whether to analyze thresholds.

        :return: An instance of ExpectedImbalanceBars.
        """
        return ExpectedImbalanceBars(
            bar_type='dollar_imbalance',
            window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
            initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
            window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
            expected_ticks_number_bounds=expected_ticks_number_bounds,
            analyse_thresholds=analyze_thresholds
        )

    @staticmethod
    def initialize_expected_volume_imbalance_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            expected_ticks_number_bounds: Tuple[float] = None,
            analyse_thresholds: bool = False,
    ):

        bars = ExpectedImbalanceBars(bar_type='volume_imbalance',
                                     window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
                                     initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
                                     window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
                                     expected_ticks_number_bounds=expected_ticks_number_bounds,
                                     analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_expected_tick_imbalance_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            expected_ticks_number_bounds: Tuple[float] = None,
            analyse_thresholds: bool = False,
    ):

        bars = ExpectedImbalanceBars(bar_type='tick_imbalance',
                                     window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
                                     initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
                                     window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
                                     expected_ticks_number_bounds=expected_ticks_number_bounds,
                                     analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_fixed_dollar_imbalance_bars(
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            analyse_thresholds: bool = False,
    ):

        bars = FixedImbalanceBars(bar_type='dollar_imbalance', window_size_for_expected_n_ticks_estimation=None,
                                  initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
                                  window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
                                  analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_fixed_volume_imbalance_bars(
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            analyse_thresholds: bool = False,
    ):

        bars = FixedImbalanceBars(bar_type='volume_imbalance', window_size_for_expected_n_ticks_estimation=None,
                                  initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
                                  window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
                                  analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_fixed_tick_imbalance_bars(
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            analyse_thresholds: bool = False,
    ):

        bars = FixedImbalanceBars(bar_type='tick_imbalance', window_size_for_expected_n_ticks_estimation=None,
                                  initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
                                  window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
                                  analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_expected_dollar_run_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            expected_ticks_number_bounds: Tuple[float] = None,
            analyse_thresholds: bool = False,
    ):

        bars = ExpectedRunBars(bar_type='dollar_run',
                               window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
                               initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
                               window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
                               expected_ticks_number_bounds=expected_ticks_number_bounds,
                               analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_expected_volume_run_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            expected_ticks_number_bounds: Tuple[float] = None,
            analyse_thresholds: bool = False,
    ):

        bars = ExpectedRunBars(bar_type='volume_run',
                               window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
                               initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
                               window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
                               expected_ticks_number_bounds=expected_ticks_number_bounds,
                               analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_expected_tick_run_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            expected_ticks_number_bounds: Tuple[float] = None,
            analyse_thresholds: bool = False,
    ):

        bars = ExpectedRunBars(bar_type='tick_run',
                               window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
                               initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
                               window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
                               expected_ticks_number_bounds=expected_ticks_number_bounds,
                               analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_fixed_dollar_run_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            analyse_thresholds: bool = False,
    ):

        bars = FixedRunBars(
            bar_type='dollar_run',
            window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
            window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
            initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
            analyse_thresholds=analyse_thresholds)

        return bars

    @staticmethod
    def initialize_fixed_volume_run_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            analyse_thresholds: bool = False,
    ):

        bars = FixedRunBars(
            bar_type='volume_run',
            window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
            window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
            initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
            analyse_thresholds=analyse_thresholds
        )

        return bars

    @staticmethod
    def initialize_fixed_tick_run_bars(
            window_size_for_expected_n_ticks_estimation: int = 3,
            window_size_for_expected_imbalance_estimation: int = 10000,
            initial_estimate_of_expected_n_ticks_in_bar: int = 20000,
            analyse_thresholds: bool = False,
    ):

        bars = FixedRunBars(
            bar_type='tick_run',
            window_size_for_expected_n_ticks_estimation=window_size_for_expected_n_ticks_estimation,
            window_size_for_expected_imbalance_estimation=window_size_for_expected_imbalance_estimation,
            initial_estimate_of_expected_n_ticks_in_bar=initial_estimate_of_expected_n_ticks_in_bar,
            analyse_thresholds=analyse_thresholds
        )

        return bars

    @staticmethod
    def initialize_dollar_standard_bars(
            threshold: Union[float, pd.Series] = 70000000,
    ):

        bars = StandardBars(
            bar_type=CUMULATIVE_DOLLAR,
            threshold=threshold,
        )
        return bars

    @staticmethod
    def initialize_volume_standard_bars(
            threshold: Union[float, pd.Series] = 30000,
    ):

        bars = StandardBars(
            bar_type=CUMULATIVE_VOLUME,
            threshold=threshold
        )

        return bars

    @staticmethod
    def initialize_tick_standard_bars(
            threshold: Union[float, pd.Series] = 6000,
    ):

        bars = StandardBars(
            bar_type=CUMULATIVE_TICKS,
            threshold=threshold,
        )
        return bars

    @staticmethod
    def initialize_time_bars(
            resolution_type: str = 'D',
            resolution_units: int = 1,
    ):

        bars = TimeBars(
            resolution_type=resolution_type,
            resolution_units=resolution_units,
        )

        return bars



import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_samples

def covariance_to_correlation_matrix(covariance: np.ndarray) -> np.ndarray:
    """
    Derive the correlation matrix from a covariance matrix.
    
    :param covariance: Covariance matrix.
    :type covariance: numpy.ndarray
    :return: Correlation matrix.
    :rtype: numpy.ndarray
    """
    standard_deviation = np.sqrt(np.diag(covariance))
    correlation = covariance / np.outer(standard_deviation, standard_deviation)
    correlation[correlation < -1] = -1
    correlation[correlation > 1] = 1
    return correlation


def get_optimal_portfolio_weights(covariance: np.ndarray, mu: np.ndarray = None) -> np.ndarray:
    """
    Compute the optimal portfolio weights.
    
    :param covariance: Covariance matrix.
    :type covariance: numpy.ndarray
    :param mu: Mean vector, defaults to None.
    :type mu: numpy.ndarray, optional
    :return: Portfolio weights.
    :rtype: numpy.ndarray
    """
    inverse_covariance = np.linalg.inv(covariance)
    ones = np.ones(shape=(inverse_covariance.shape[0], 1))
    if mu is None:
        mu = ones
    weights = np.dot(inverse_covariance, mu)
    weights /= np.dot(ones.T, weights)
    return weights


def get_optimal_portfolio_weights_nco(
        covariance: np.ndarray,
        mu: np.ndarray = None,
        number_clusters: int = None) -> np.ndarray:
    """
    Compute the optimal portfolio weights using the NCO algorithm.
    
    :param covariance: Covariance matrix.
    :type covariance: numpy.ndarray
    :param mu: Mean vector, defaults to None.
    :type mu: numpy.ndarray, optional
    :param number_clusters: Maximum number of clusters, defaults to None.
    :type number_clusters: int, optional
    :return: Optimal portfolio weights using NCO algorithm.
    :rtype: numpy.ndarray
    """
    covariance = pd.DataFrame(covariance)
    correlation = covariance_to_correlation_matrix(covariance)
    if mu is not None:
        mu = pd.Series(mu[:, 0])
    if number_clusters is None:
        number_clusters = int(correlation.shape[0] / 2)
    correlation, clusters, _ = cluster_k_means_base(correlation, number_clusters)
    weights_intra_cluster = pd.DataFrame(0, index=covariance.index, columns=clusters.keys())
    for i in clusters:
        covariance_intra_cluster = covariance.loc[clusters[i], clusters[i]].values
        if mu is None:
            mu_intra_cluster = None
        else:
            mu_intra_cluster = mu.loc[clusters[i]].values.reshape(-1, 1)
        weights_intra_cluster.loc[clusters[i], i] = get_optimal_portfolio_weights(
            covariance_intra_cluster, mu_intra_cluster).flatten()
    covariance_inter_cluster = weights_intra_cluster.T.dot(np.dot(covariance, weights_intra_cluster))
    mu_inter_cluster = None if mu is None else weights_intra_cluster.T.dot(mu)
    weights_inter_cluster = pd.Series(
        get_optimal_portfolio_weights(covariance_inter_cluster, mu_inter_cluster).flatten(),
        index=covariance_inter_cluster.index)
    weights_nco = weights_intra_cluster.mul(weights_inter_cluster, axis=1).sum(axis=1).values.reshape(-1, 1)
    return weights_nco


def cluster_k_means_base(
        correlation: pd.DataFrame,
        number_clusters: int = 10,
        iterations: int = 10) -> tuple:
    """
    Perform clustering using the K-means algorithm.
    
    :param correlation: Correlation matrix.
    :type correlation: pd.DataFrame
    :param number_clusters: Maximum number of clusters, defaults to 10.
    :type number_clusters: int, optional
    :param iterations: Number of iterations, defaults to 10.
    :type iterations: int, optional
    :return: Updated correlation matrix, cluster members, silhouette scores.
    :rtype: tuple
    """
    distance = ((1 - correlation.fillna(0)) / 2.0) ** 0.5
    silhouette_scores = pd.Series()

    for init in range(iterations):
        for i in range(2, number_clusters + 1):
            kmeans_ = KMeans(n_clusters=i, n_init=1)
            kmeans_ = kmeans_.fit(distance)
            silhouette_scores_ = silhouette_samples(distance, kmeans_.labels_)
            statistic = (silhouette_scores_.mean() / silhouette_scores_.std(),
                         silhouette_scores.mean() / silhouette_scores.std())
            if np.isnan(statistic[1]) or statistic[0] > statistic[1]:
                silhouette_scores, kmeans = silhouette_scores_, kmeans_

    index_new = np.argsort(kmeans.labels_)
    correlation_new = correlation.iloc[index_new]
    correlation_new = correlation_new.iloc[:, index_new]
    clusters = {i: correlation.columns[np.where(kmeans.labels_ == i)[0]].tolist() for i in np.unique(kmeans.labels_)}
    silhouette_scores = pd.Series(silhouette_scores, index=distance.index)
 
    return correlation_new, clusters, silhouette_scores



import pandas as pd
import numpy as np 
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor

def cusum_filter_events_dynamic_threshold(
        prices: pd.Series,
        threshold: pd.Series
) -> pd.DatetimeIndex:
    """
    Detect events using the Symmetric Cumulative Sum (CUSUM) filter.

    The Symmetric CUSUM filter is a change-point detection algorithm used to identify events where the price difference
    exceeds a predefined threshold.

    :param prices: A pandas Series of prices.
    :param threshold: A pandas Series containing the predefined threshold values for event detection.
    :return: A pandas DatetimeIndex containing timestamps of detected events.

    References:
    - De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons. (Methodology: 39)
    """
    time_events, shift_positive, shift_negative = [], 0, 0
    price_delta = prices.diff().dropna()
    thresholds = threshold.copy()
    price_delta, thresholds = price_delta.align(thresholds, join="inner", copy=False)

    for (index, value), threshold_ in zip(price_delta.to_dict().items(), thresholds.to_dict().values()):
        shift_positive = max(0, shift_positive + value)
        shift_negative = min(0, shift_negative + value)

        if shift_negative < -threshold_:
            shift_negative = 0
            time_events.append(index)

        elif shift_positive > threshold_:
            shift_positive = 0
            time_events.append(index)

    return pd.DatetimeIndex(time_events)

def symmetric_cusum_filter(
        prices: pd.Series,
        threshold: float) -> pd.DatetimeIndex:
    """
    Implements the symmetric CUSUM filter.

    The symmetric CUSUM filter is a change-point detection algorithm used to identify events where the price difference exceeds a predefined threshold.

    :param prices: A pandas Series of prices.
    :param threshold: The predefined threshold for detecting events.
    :return: A pandas DatetimeIndex of event timestamps.

    References:
    - De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons. (Methodology: 39)
    """
    time_events, shift_positive, shift_negative = [], 0, 0
    price_delta = prices.diff()

    for i in price_delta.index[1:]:
        shift_positive = max(0, shift_positive + price_delta.loc[i])
        shift_negative = min(0, shift_negative + price_delta.loc[i])

        if shift_negative < -threshold:
            shift_negative = 0
            time_events.append(i)

        elif shift_positive > threshold:
            shift_positive = 0
            time_events.append(i)

    return pd.DatetimeIndex(time_events)




def aggregate_ohlcv(tick_data_grouped) -> pd.DataFrame:
    """
    Aggregates tick data into OHLCV bars.

    :param tick_data_grouped: A pandas GroupBy object of tick data.
    :return: A pandas DataFrame with OHLCV bars.
    """
    ohlc = tick_data_grouped['price'].ohlc()
    ohlc['volume'] = tick_data_grouped['size'].sum()
    ohlc['value_of_trades'] = tick_data_grouped.apply(
        lambda x: (x['price'] * x['size']).sum() / x['size'].sum())
    ohlc['price_mean'] = tick_data_grouped['price'].mean()
    ohlc['tick_count'] = tick_data_grouped['price'].count()
    ohlc['price_mean_log_return'] = np.log(ohlc['price_mean']) - np.log(ohlc['price_mean'].shift(1))

    return ohlc


def generate_time_bars(
        tick_data: pd.DataFrame,
        frequency: str = "5Min") -> pd.DataFrame:
    """
    Generates time bars from tick data.

    :param tick_data: A pandas DataFrame of tick data.
    :param frequency: The frequency for time bar aggregation.
    :return: A pandas DataFrame with time bars.
    """
    tick_data_grouped = tick_data.groupby(pd.Grouper(freq=frequency))
    ohlcv_dataframe = aggregate_ohlcv(tick_data_grouped)

    return ohlcv_dataframe


def compute_daily_volatility(
        close: pd.Series,
        span: int = 63) -> pd.DataFrame:
    """
    Computes the daily volatility at intraday estimation points.

    :param close: A pandas Series of close prices.
    :param span: The span parameter for the EWMA.
    :return: A pandas DataFrame with returns and volatilities.

    References:
    - De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons. (Methodology: Page 44)
    """
    df = close.index.searchsorted(close.index - pd.Timedelta(days=1))
    df = df[df > 0]
    df = pd.Series(close.index[df - 1], index=close.index[close.shape[0] - df.shape[0]:])
    returns = (close.loc[df.index] / close.loc[df.values].values - 1).rename("rets")
    stds = returns.ewm(span=span).std().rename("std")

    return pd.concat([returns, stds], axis=1)

def daily_volatility_with_log_returns(
        close: pd.Series,
        span: int = 100
) -> pd.Series:
    """
    Calculate the daily volatility at intraday estimation points using Exponentially Weighted Moving Average (EWMA).

    :param close: A pandas Series of daily close prices.
    :param span: The span parameter for the Exponentially Weighted Moving Average (EWMA).
    :return: A pandas Series containing daily volatilities.

    References:
    - De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons. (Methodology: Page 44)
    """
    df = close.index.searchsorted(close.index - pd.Timedelta(days=1))
    df = df[df > 0]
    df = pd.Series(close.index[df - 1], index=close.index[close.shape[0] - df.shape[0]:])
    returns = np.log(close.loc[df.index] / close.loc[df.values].values)
    stds = returns.ewm(span=span).std().rename("std")

    return stds

def triple_barrier(
    close: pd.Series,
    events: pd.DataFrame,
    profit_taking_stop_loss: list[float, float],
    molecule: list
) -> pd.DataFrame:
    """
    Implements the triple-barrier method.

    :param close: A pandas Series of close prices.
    :param events: A pandas DataFrame with the timestamp of the vertical barrier and the unit width of the horizontal barriers.
    :param profit_taking_stop_loss: A list of two non-negative float values to set the width of the upper and lower barrier.
    :param molecule: A list with the subset of event indices.
    :return: A pandas DataFrame with the vertical barrier and two horizontal barriers.

    References:
    - De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons. (Methodology: Page 45)
    """
    events_filtered = events.loc[molecule]
    output = events_filtered[['End Time']].copy(deep=True)

    if profit_taking_stop_loss[0] > 0:
        profit_taking = profit_taking_stop_loss[0] * events_filtered['Base Width']
    else:
        profit_taking = pd.Series(index=events.index)

    if profit_taking_stop_loss[1] > 0:
        stop_loss = -profit_taking_stop_loss[1] * events_filtered['Base Width']
    else:
        stop_loss = pd.Series(index=events.index)

    for location, timestamp in events_filtered['End Time'].fillna(close.index[-1]).items():
        df = close[location:timestamp]
        df = np.log(df / close[location]) * events_filtered.at[location, 'Side']
        output.loc[location, 'stop_loss'] = df[df < stop_loss[location]].index.min()
        output.loc[location, 'profit_taking'] = df[df > profit_taking[location]].index.min()

    return output

def get_barrier_touch_time(close: pd.Series, 
                          time_events: pd.DatetimeIndex, 
                          ptsl: float, 
                          target: pd.Series, 
                          return_min: float, 
                          num_threads: int, 
                          timestamp: pd.Series = False) -> pd.DataFrame:
    """
    Finds the time of the first barrier touch.
    
    :param close: A dataframe of dates and close prices.
    :param time_events: A pandas time index containing the timestamps that will seed every triple barrier.
    :param ptsl: A non-negative float that sets the width of the two barriers.
    :param target: A pandas series of targets, expressed in terms of absolute returns.
    :param return_min: The minimum target return required for running a triple barrier search.
    :param num_threads: The number of threads.
    :param timestamp: A pandas series with the timestamps of the vertical barriers (False when disabled).
    :return: A dataframe with timestamp of the vertical barrier and unit width of the horizontal barriers.
    """
    target = target.loc[time_events]
    target = target[target > return_min]

    if timestamp is False:
        timestamp = pd.Series(pd.NaT, index=time_events)

    side_position = pd.Series(1., index=target.index)
    events = pd.concat({'timestamp': timestamp, 'target': target, 'side': side_position}, axis=1).dropna(subset=['target'])
    
    with ProcessPoolExecutor(num_threads) as executor:
        dataframe = list(executor.map(triple_barrier, [(close, events.loc[molecule], [ptsl, ptsl]) for molecule in np.array_split(events.index, num_threads)]))
    dataframe = pd.concat(dataframe, axis=0)
    
    events['timestamp'] = dataframe.dropna(how='all').min(axis=1)
    events = events.drop('side', axis=1)
    return events


def vertical_barrier(
    close: pd.Series,
    time_events: pd.DatetimeIndex,
    number_days: int
) -> pd.Series:
    """
    Shows one way to define a vertical barrier.

    :param close: A dataframe of prices and dates.
    :param time_events: A vector of timestamps.
    :param number_days: A number of days for the vertical barrier.
    :return: A pandas series with the timestamps of the vertical barriers.
    """
    timestamp_array = close.index.searchsorted(time_events + pd.Timedelta(days=number_days))
    timestamp_array = timestamp_array[timestamp_array < close.shape[0]]
    timestamp_array = pd.Series(close.index[timestamp_array], index=time_events[:timestamp_array.shape[0]])
    return timestamp_array


def get_labels(events: pd.DataFrame, 
               close: pd.Series) -> pd.DataFrame:
    """
    Label the observations.
    
    :param events: A dataframe with timestamp of the vertical barrier and unit width of the horizontal barriers.
    :param close: A dataframe of dates and close prices.
    :return: A dataframe with the return realized at the time of the first touched barrier and the label.
    """
    events_filtered = events.dropna(subset=['timestamp'])
    all_dates = events_filtered.index.union(events_filtered['timestamp'].values).drop_duplicates()
    close_filtered = close.reindex(all_dates, method='bfill')
    
    out = pd.DataFrame(index=events_filtered.index)
    out['ret'] = close_filtered.loc[events_filtered['timestamp'].values].values / close_filtered.loc[events_filtered.index] - 1
    out['bin'] = np.sign(out['ret'])
    return out

def meta_events(
    close: pd.Series,
    time_events: pd.DatetimeIndex,
    ptsl: list,
    target: pd.Series,
    return_min: float,
    num_threads: int,
    timestamp: pd.Series = False,
    side: pd.Series = None
) -> pd.DataFrame:
    """
    Expand events to incorporate meta-labeling.

    :param close: A dataframe of prices and dates.
    :param time_events: A vector of timestamps.
    :param ptsl: A list of two non-negative float values that multiply the target.
    :param target: A dataframe of targets, expressed in terms of absolute returns.
    :param return_min: The minimum target return required for running a triple barrier search.
    :param num_threads: The number of threads.
    :param timestamp: A vector containing the timestamps of the vertical barriers (False when disabled).
    :param side: If not None, the function understands that meta-labeling is in play.
    :return: A dataframe with timestamp of the vertical barrier and unit width of the horizontal barriers.
    """
    target = target.loc[time_events]
    target = target[target > return_min]

    if timestamp is False:
        timestamp = pd.Series(pd.NaT, index=time_events)

    if side is None:
        side_position, profit_loss = pd.Series(1., index=target.index), [ptsl[0], ptsl[0]]
    else:
        side_position, profit_loss = side.loc[target.index], ptsl[:2]

    events = pd.concat({'End Time': timestamp, 'Base Width': target, 'Side': side_position}, axis=1).dropna(subset=['Base Width'])
    if num_threads > 1:
        with ProcessPoolExecutor(num_threads) as executor:
            df0 = list(executor.map(
                triple_barrier,
                [close] * num_threads,
                [events] * num_threads,
                [profit_loss] * num_threads,
                list(np.array_split(time_events, num_threads))
            ))
    else:
        df0 = list(map(
            triple_barrier,
            [close] * num_threads,
            [events] * num_threads,
            [profit_loss] * num_threads,
            list(np.array_split(time_events, num_threads))
        ))        
    df0 = pd.concat(df0, axis=0)

    events['End Time'] = df0.dropna(how='all').min(axis=1)

    if side is None:
        events = events.drop('Side', axis=1)
    return events
        
import numpy as np
import pandas as pd
import multiprocessing as mp
import datetime
import time
import sys

def meta_labeling(
    events: pd.DataFrame,
    close: pd.Series
) -> pd.DataFrame:
    """
    Expands label to incorporate meta-labeling.

    :param events: DataFrame with timestamp of vertical barrier and unit width of the horizontal barriers.
    :param close: Series of close prices with date indices.
    :return: DataFrame containing the return and binary labels for each event.

    Reference:
    De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons.
    Methodology: 51
    """
    events_filtered = events.dropna(subset=['End Time'])
    all_dates = events_filtered.index.union(events_filtered['End Time'].values).drop_duplicates()
    close_filtered = close.reindex(all_dates, method='bfill')
    out = pd.DataFrame(index=events_filtered.index)
    out['End Time'] = events['End Time']
    out['Return'] = close_filtered.loc[events_filtered['End Time'].values].values / close_filtered.loc[events_filtered.index] - 1
    if 'Side' in events_filtered:
        out['Return'] *= events_filtered['Side']
    out['Label'] = np.sign(out['Return'])
    if 'Side' in events_filtered:
        out.loc[out['Return'] <= 0, 'Label'] = 0
        out['Side'] = events_filtered['Side']
    return out

def drop_label(
        events: pd.DataFrame,
        percent_min: float = .05) -> pd.DataFrame:
    """
    Presents a procedure that recursively drops observations associated with extremely rare labels.

    :param events: DataFrame with columns: Dates, ret, and bin.
    :param percent_min: Minimum percentage.
    :return: DataFrame with the updated events.

    Reference:
    De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons.
    Methodology: 54
    """
    while True:
        dataframe = events['bin'].value_counts(normalize=True)
        if dataframe.min() > percent_min or dataframe.shape[0] < 3:
            break
        print('dropped label', dataframe.idxmin(), dataframe.min())
        events = events[events['bin'] != dataframe.idxmin()]
    return events


def lin_parts(num_atoms: int, num_threads: int) -> np.ndarray:
    """
    Partition of atoms with a single loop.

    :param num_atoms: Total number of atoms.
    :param num_threads: Number of threads for parallel processing.
    :return: Numpy array with partition indices.
    """
    parts = np.linspace(0, num_atoms, min(num_threads, num_atoms) + 1)
    parts = np.ceil(parts).astype(int)
    return parts


def nested_parts(num_atoms: int, num_threads: int, upper_triang: bool = False) -> np.ndarray:
    """
    Partition of atoms with an inner loop.

    :param num_atoms: Total number of atoms.
    :param num_threads: Number of threads for parallel processing.
    :param upper_triang: Whether the first rows are the heaviest.
    :return: Numpy array with partition indices.
    """
    parts, num_threads_ = [0], min(num_threads, num_atoms)
    for num in range(num_threads_):
        part = 1 + 4 * (parts[-1] ** 2 + parts[-1] + num_atoms * (num_atoms + 1.) / num_threads_)
        part = (-1 + part ** .5) / 2.
        parts.append(part)
    parts = np.round(parts).astype(int)
    if upper_triang:
        parts = np.cumsum(np.diff(parts)[::-1])
        parts = np.append(np.array([0]), parts)
    return parts


def mp_pandas_obj(
        func,
        pd_obj,
        num_threads: int = 24,
        mp_batches: int = 1,
        lin_mols: bool = True,
        **kargs) -> pd.DataFrame:
    """
    Parallelize jobs, return a DataFrame or Series.

    :param func: Function to be parallelized.
    :param pd_obj: Tuple with argument name for the molecule and list of atoms grouped into molecules.
    :param num_threads: Number of threads for parallel processing.
    :param mp_batches: Number of multi-processing batches.
    :param lin_mols: Whether to use linear molecule partitioning.
    :param kargs: Any other arguments needed by func.
    :return: DataFrame with the results of the parallelized function.

    Example:
    df1 = mp_pandas_obj(func, ('molecule', df0.index), 24, **kargs)
    """
    arg_list = list(kargs.values())
    if lin_mols:
        parts = lin_parts(len(arg_list[1]), num_threads * mp_batches)
    else:
        parts = nested_parts(len(arg_list[1]), num_threads * mp_batches)
    jobs = []
    for i in range(1, len(parts)):
        job = {pd_obj[0]: pd_obj[1][parts[i - 1]:parts[i]], 'func': func}
        job.update(kargs)
        jobs.append(job)
    if num_threads == 1:
        out = process_jobs_(jobs)
    else:
        out = process_jobs(jobs, num_threads=num_threads)
    if isinstance(out[0], pd.DataFrame):
        df0 = pd.DataFrame()
    elif isinstance(out[0], pd.Series):
        df0 = pd.Series()
    else:
        return out
    for i in out:
        df0 = df0.append(i)
    df0 = df0.sort_index()
    return df0


def process_jobs_(jobs: list) -> list:
    """
    Run jobs sequentially, for debugging.

    :param jobs: List of jobs to be processed.
    :return: List of job results.
    """
    out = []
    for job in jobs:
        out_ = expand_call(job)
        out.append(out_)
    return out


def report_progress(job_num: int, num_jobs: int, time0: float, task: str) -> None:
    """
    Report progress as asynchronous jobs are completed.

    :param job_num: Current job number.
    :param num_jobs: Total number of jobs.
    :param time0: Start time.
    :param task: Task name.
    :return: None
    """
    msg = [float(job_num) / num_jobs, (time.time() - time0) / 60.]
    msg.append(msg[1] * (1 / msg[0] - 1))
    timestamp = str(datetime.datetime.fromtimestamp(time.time()))
    msg = timestamp + ' ' + str(round(msg[0] * 100, 2)) + '% ' + task + ' done after ' + str(
        round(msg[1], 2)) + ' minutes. Remaining ' + str(round(msg[2], 2)) + ' minutes.'
    if job_num < num_jobs:
        sys.stderr.write(msg + '\r')
    else:
        sys.stderr.write(msg + '\n')
    return


def process_jobs(jobs: list, task: str = None, num_threads: int = 24) -> list:
    """
    Run jobs in parallel in multiple threads.

    :param jobs: List of jobs to be processed.
    :param task: Task name for progress reporting.
    :param num_threads: Number of threads for parallel processing.
    :return: List of job results.
    """
    if task is None:
        task = jobs[0]['func'].__name__
    pool = mp.Pool(processes=num_threads)
    outputs, out, time0 = pool.imap_unordered(expand_call, jobs), [], time.time()
    # Process asyn output, report progress
    for i, out_ in enumerate(outputs, 1):
        out.append(out_)
        report_progress(i, len(jobs), time0, task)
    pool.close()
    pool.join()  # this is needed to prevent memory leaks
    return out


def expand_call(kargs):
    """
    Expand the arguments of a callback function, kargs['func'].

    :param kargs: Dictionary with the function to call and the arguments to pass.
    :return: Result of the function call.
    """
    func = kargs['func']
    del kargs['func']
    out = func(**kargs)
    return out



from typing import Union, Dict, Any
import pandas as pd
import numpy as np
from collections import ChainMap
from joblib import Parallel, delayed
import warnings
from sklearn.exceptions import ConvergenceWarning
from sklearn.ensemble import BaggingClassifier, BaggingRegressor
from joblib_progress import joblib_progress
from sklearn.base import clone

from .combinatorial_purged import CombinatorialPurged

class BaggedCombinatorialPurged(CombinatorialPurged):
    def __init__(
        self,
        n_splits: int,
        n_test_groups: int,
        times: Union[pd.Series, Dict[str, pd.Series]],
        embargo: float = 0,
        classifier: bool = True,
        n_estimators: int = 10,
        max_samples: float = 1.0,
        max_features: float = 1.0,
        bootstrap: bool = True,
        bootstrap_features: bool = False,
        random_state: int = None 
    ):
        """
        Initialize the BaggedCombinatorialPurged class.

        Parameters
        ----------
        n_splits : int
            Number of splits/groups to partition the data into.
        n_test_groups : int
            Size of the testing set in terms of groups.
        times : Union[pd.Series, Dict[str, pd.Series]]
            The timestamp series associated with the labels.
        embargo : float
            The embargo rate for purging.
        classifier : bool
            Determines whether to use a BaggingClassifier or BaggingRegressor.
        n_estimators : int
            The number of base estimators in the ensemble.
        max_samples : float
            The number of samples to draw from X to train each base estimator.
        max_features : float
            The number of features to draw from X to train each base estimator.
        bootstrap : bool
            Whether samples are drawn with replacement.
        bootstrap_features : bool
            Whether features are drawn with replacement.
        random_state : int
            The seed used by the random number generator.
        """
        super().__init__(n_splits, n_test_groups, times, embargo)
        self.classifier = classifier
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.bootstrap_features = bootstrap_features
        self.random_state = random_state

    def _single_backtest_predictions(
        self,
        single_estimator: Any,
        single_times: pd.Series,
        single_data: pd.DataFrame,
        single_labels: pd.Series,
        single_weights: np.ndarray,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Dict[str, np.ndarray]:
        """
        Generate predictions for a single backtest using combinatorial splits with bagging.

        This method calculates predictions across various paths created by combinatorial splits
        of the data. For each combinatorial split, a bagged estimator is trained and then used
        to predict on the corresponding test set.

        :param single_estimator: The machine learning model or estimator to be trained.
        :param single_times: Timestamps corresponding to the data points.
        :param single_data: Input data on which the model is trained and predictions are made.
        :param single_labels: Labels corresponding to the input data.
        :param single_weights: Weights for each data point.
        :param predict_probability: If True, predict the probability of forecasts.
        :type predict_probability: bool
        :param n_jobs: Number of CPU cores to use for parallelization. Default is 1.

        :return: A dictionary where keys are path names and values are arrays of predictions.

        .. note:: This function relies on internal methods (e.g., `_get_train_indices`)
                to manage data splits and training.

        .. note:: Parallelization is used to speed up the training of models for different splits.
        """
        
        self._validate_input(single_times, single_data)

        if single_weights is None:
            single_weights = pd.Series(np.ones((len(single_data),)))

        paths_predictions = {}
        combinations_, locations, split_segments = self._combinations_and_path_locations_and_split_segments(single_data)

        def train_single_estimator(
            single_estimator_,
            combinatorial_test_indices
        ):
            combinatorial_train_indices = self._get_train_indices(combinatorial_test_indices, single_times)

            X_train = single_data.iloc[combinatorial_train_indices]
            y_train = single_labels.iloc[combinatorial_train_indices]
            weights_train = single_weights.iloc[combinatorial_train_indices]

            if self.classifier:
                bagging_estimator = BaggingClassifier(
                    estimator=single_estimator_,
                    n_estimators=self.n_estimators,
                    max_samples=self.max_samples,
                    max_features=self.max_features,
                    bootstrap=self.bootstrap,
                    bootstrap_features=self.bootstrap_features,
                    random_state=self.random_state
                )
            else:
                bagging_estimator = BaggingRegressor(
                    estimator=single_estimator_,
                    n_estimators=self.n_estimators,
                    max_samples=self.max_samples,
                    max_features=self.max_features,
                    bootstrap=self.bootstrap,
                    bootstrap_features=self.bootstrap_features,
                    random_state=self.random_state
                )

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=ConvergenceWarning)
                try:
                    bagging_estimator.fit(X_train, y_train, sample_weight=weights_train)

                except (TypeError, ValueError):
                    bagging_estimator.fit(X_train, y_train)

            return bagging_estimator

        # with joblib_progress("Backtesting...", total=self.get_n_splits()): 
        combinatorial_trained_estimators = Parallel(n_jobs=n_jobs)(
            delayed(train_single_estimator)(clone(single_estimator), combinatorial_test_indices)
            for combinatorial_test_indices in CombinatorialPurged._combinatorial_splits(combinations_, split_segments))


        def get_path_data(
            path_num,
            locs
        ):
            path_data = []

            for (G, S) in locs:
                test_indices = split_segments[G]
                X_test = single_data.iloc[test_indices]

                if predict_probability:
                    predictions = combinatorial_trained_estimators[S].predict_proba(X_test)

                else:
                    predictions = combinatorial_trained_estimators[S].predict(X_test)

                path_data.extend(predictions)

            path_data = {f"Path {path_num}" : np.array(path_data)}

            return path_data

        paths_predictions = Parallel(n_jobs=n_jobs)(delayed(get_path_data)(path_num, locs) for path_num, locs in locations.items())
        paths_predictions = dict(ChainMap(*reversed(paths_predictions)))

        return paths_predictions



from .combinatorial_purged import *
from .cross_validator_controller import *
from .cross_validator_factory import *
from .cross_validator_interface import *
from .kfold import *
from .purged_kfold import *
from .walk_forward import *
from .adaptive_combinatorial_purged import *
from .bagged_combinatorial_purged import *



from RiskLabAI.data.structures.abstract_information_driven_bars import *
from RiskLabAI.utils.constants import *


class AbstractRunBars(AbstractInformationDrivenBars):
    """
    Abstract class that contains the run properties which are shared between the subtypes.
    This class subtypes are as follows:
        1- ExpectedRunBars
        2- FixedRunBars

    The class implements run bars sampling logic as explained on page 31,32 of Advances in Financial Machine Learning.
    """

    def __init__(
        self,
        bar_type: str,
        window_size_for_expected_n_ticks_estimation: int,
        window_size_for_expected_imbalance_estimation: int,
        initial_estimate_of_expected_n_ticks_in_bar: int,
        analyse_thresholds: bool
    ):
        """
        AbstractRunBars constructor function
        :param bar_type: type of bar. e.g. expected_dollar_run_bars, fixed_tick_run_bars etc.
        :param window_size_for_expected_n_ticks_estimation: window size used to estimate number of ticks expectation
        :param initial_estimate_of_expected_n_ticks_in_bar: initial estimate of number of ticks expectation window size
        :param window_size_for_expected_imbalance_estimation: window size used to estimate imbalance expectation
        :param analyse_thresholds: whether return thresholds values (θ, number of ticks expectation, imbalance expectation) in a tabular format
        """
        super().__init__(
            bar_type,
            window_size_for_expected_n_ticks_estimation,
            initial_estimate_of_expected_n_ticks_in_bar,
            window_size_for_expected_imbalance_estimation
        )

        self.run_bars_statistics = {
            CUMULATIVE_BUY_θ: 0,
            CUMULATIVE_SELL_θ: 0,

            EXPECTED_BUY_IMBALANCE: np.nan,
            EXPECTED_SELL_IMBALANCE: np.nan,
            EXPECTED_BUY_TICKS_PROPORTION: np.nan,

            BUY_TICKS_NUMBER: 0,

            # Previous bars number of ticks and previous tick imbalances
            PREVIOUS_BARS_N_TICKS_LIST: [],
            PREVIOUS_TICK_IMBALANCES_BUY_LIST: [],
            PREVIOUS_TICK_IMBALANCES_SELL_LIST: [],
            PREVIOUS_BARS_BUY_TICKS_PROPORTIONS_LIST: []
        }

        if analyse_thresholds:
            self.analyse_thresholds = []
        else:
            self.analyse_thresholds = None

    def construct_bars_from_data(self, data: Union[list, tuple, np.ndarray]) -> list:
        """
        The function is used to construct bars from input ticks data.
        :param data: tabular data that contains date_time, price, and volume columns
        :return: constructed bars
        """
        bars_list = []
        for tick_data in data:
            self.tick_counter += 1

            date_time, price, volume = tuple(tick_data)
            tick_rule = self._tick_rule(price)
            self.update_base_fields(price, tick_rule, volume)

            # imbalance calculations
            imbalance = self._imbalance_at_tick(price, tick_rule, volume)

            if imbalance > 0:
                self.run_bars_statistics[PREVIOUS_TICK_IMBALANCES_BUY_LIST].append(imbalance)
                self.run_bars_statistics[CUMULATIVE_BUY_θ] += imbalance
                self.run_bars_statistics[BUY_TICKS_NUMBER] += 1

            elif imbalance < 0:
                self.run_bars_statistics[PREVIOUS_TICK_IMBALANCES_SELL_LIST].append(-imbalance)
                self.run_bars_statistics[CUMULATIVE_SELL_θ] += -imbalance

            warm_up = np.isnan([self.run_bars_statistics[EXPECTED_BUY_IMBALANCE],
                                self.run_bars_statistics[EXPECTED_SELL_IMBALANCE]]).any()

            # initialize expected imbalance first time, when initial_estimate_of_expected_n_ticks_in_bar passed
            if len(bars_list) == 0 and warm_up:

                self.run_bars_statistics[EXPECTED_BUY_IMBALANCE] = self._ewma_expected_imbalance(
                    self.run_bars_statistics[PREVIOUS_TICK_IMBALANCES_BUY_LIST],
                    self.information_driven_bars_statistics[EXPECTED_IMBALANCE_WINDOW],
                    warm_up
                )
                self.run_bars_statistics[EXPECTED_SELL_IMBALANCE] = self._ewma_expected_imbalance(
                    self.run_bars_statistics[PREVIOUS_TICK_IMBALANCES_SELL_LIST],
                    self.information_driven_bars_statistics[EXPECTED_IMBALANCE_WINDOW],
                    warm_up
                )

                if not np.isnan([self.run_bars_statistics[EXPECTED_BUY_IMBALANCE],
                                 self.run_bars_statistics[EXPECTED_SELL_IMBALANCE]]).any():
                    self.run_bars_statistics[EXPECTED_BUY_TICKS_PROPORTION] = \
                        self.run_bars_statistics[BUY_TICKS_NUMBER] / self.base_statistics[CUMULATIVE_TICKS]

            if self.analyse_thresholds is not None:
                self.run_bars_statistics[TIMESTAMP] = date_time
                self.analyse_thresholds.append(dict(self.run_bars_statistics))

            # is construction condition met to construct next bar or not

            max_proportion = max(
                self.information_driven_bars_statistics[EXPECTED_IMBALANCE_WINDOW] *
                self.run_bars_statistics[EXPECTED_BUY_TICKS_PROPORTION],

                self.run_bars_statistics[EXPECTED_SELL_IMBALANCE] *
                (1 - self.run_bars_statistics[EXPECTED_BUY_TICKS_PROPORTION])
            )
            threshold = self.information_driven_bars_statistics[EXPECTED_TICKS_NUMBER] * max_proportion
            is_construction_condition_met = self._bar_construction_condition(threshold)
            if is_construction_condition_met:
                next_bar = self._construct_next_bar(
                    date_time,
                    self.tick_counter,
                    price,
                    self.high_price,
                    self.low_price,
                    threshold
                )

                bars_list.append(next_bar)

                self.run_bars_statistics[PREVIOUS_BARS_N_TICKS_LIST].append(self.base_statistics[CUMULATIVE_TICKS])
                self.run_bars_statistics[PREVIOUS_BARS_BUY_TICKS_PROPORTIONS_LIST].append(
                    self.run_bars_statistics[BUY_TICKS_NUMBER] / self.base_statistics[CUMULATIVE_TICKS])

                # update expected number of ticks based on formed bars
                self.information_driven_bars_statistics[EXPECTED_TICKS_NUMBER] = self._expected_number_of_ticks()

                # update buy ticks proportions
                self.run_bars_statistics[EXPECTED_BUY_TICKS_PROPORTION] = ewma(
                    np.array(self.run_bars_statistics[PREVIOUS_BARS_BUY_TICKS_PROPORTIONS_LIST][
                             -self.window_size_for_expected_n_ticks_estimation:],
                             dtype=float),
                    self.window_size_for_expected_n_ticks_estimation)[-1]

                # update expected imbalance (ewma)
                self.run_bars_statistics[EXPECTED_BUY_IMBALANCE] = self._ewma_expected_imbalance(
                    self.run_bars_statistics[PREVIOUS_TICK_IMBALANCES_BUY_LIST],
                    self.information_driven_bars_statistics[EXPECTED_IMBALANCE_WINDOW]
                )
                self.run_bars_statistics[EXPECTED_SELL_IMBALANCE] = self._ewma_expected_imbalance(
                    self.run_bars_statistics[PREVIOUS_TICK_IMBALANCES_SELL_LIST],
                    self.information_driven_bars_statistics[EXPECTED_IMBALANCE_WINDOW]
                )

                # reset cached fields
                self._reset_cached_fields()

        return bars_list

    def _bar_construction_condition(
            self,
            threshold
    ):
        """
        Compute the condition of whether next bar should sample with current and previous tick datas or not.
        :return: whether next bar should form with current and previous tick datas or not.
        """

        max_proportion = max(
            self.information_driven_bars_statistics[EXPECTED_IMBALANCE_WINDOW] *
            self.run_bars_statistics[EXPECTED_BUY_TICKS_PROPORTION],

            self.run_bars_statistics[EXPECTED_SELL_IMBALANCE] *
            (1 - self.run_bars_statistics[EXPECTED_BUY_TICKS_PROPORTION])
        )

        if not np.isnan(max_proportion):
            max_θ = max(
                self.run_bars_statistics[CUMULATIVE_BUY_θ],
                self.run_bars_statistics[CUMULATIVE_SELL_θ]
            )

            condition_is_met = max_θ > threshold
            return condition_is_met
        else:
            return False

    def _reset_cached_fields(self):
        """
        This function are used (directly or override) by all concrete or abstract subtypes. The function is used to reset cached fields in bars construction process when next bar is sampled.
        :return:
        """
        super()._reset_cached_fields()
        self.run_bars_statistics[CUMULATIVE_BUY_θ] = 0
        self.run_bars_statistics[CUMULATIVE_SELL_θ] = 0
        self.run_bars_statistics[BUY_TICKS_NUMBER] = 0

    @abstractmethod
    def _expected_number_of_ticks(self) -> Union[float, int]:
        """
        Calculate number of ticks expectation when new imbalance bar is sampled.
        """


from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import KFold
from sklearn.metrics import log_loss
# from feature_importance_strategy import FeatureImportanceStrategy

from RiskLabAI.features.feature_importance.feature_importance_strategy import FeatureImportanceStrategy
from typing import Dict, Tuple, List
import numpy as np
import pandas as pd


class ClusteredFeatureImportanceMDA(FeatureImportanceStrategy):

    def __init__():
        pass

    def compute(
            self,
            classifier: RandomForestClassifier,
            x: pd.DataFrame,
            y: pd.Series,
            clusters: Dict[str, List[str]],
            n_splits: int = 10,
            score_sample_weights: List[float] = None,
            train_sample_weights: List[float] = None
    ) -> pd.DataFrame:
        """
        Compute clustered feature importance using MDA.

        The feature importance is computed by comparing the performance
        (log loss) of a trained classifier on shuffled data to its 
        performance on non-shuffled data.

        :param classifier: The Random Forest classifier to be trained.
        :param x: The features DataFrame.
        :param y: The target Series.
        :param clusters: A dictionary where the keys are the cluster names
                         and the values are lists of features in each cluster.
        :param n_splits: The number of splits for KFold cross-validation.
        :param score_sample_weights: Sample weights to be used when computing the score.
        :param train_sample_weights: Sample weights to be used during training.
        
        :return: A DataFrame with feature importances and their standard deviations.

        The related mathematical formulae:
        
        .. math::

            \\text{{importance}} = \\frac{{-1 \\times \\text{{score with shuffled data}}}}
                                  {{\\text{{score without shuffled data}}}}
        
        Using Central Limit Theorem for calculating the standard deviation:

        .. math::

            \\text{{StandardDeviation}} = \\text{{std}} \\times n^{-0.5}

        """

        # Handle default values for sample weights
        if train_sample_weights is None:
            train_sample_weights = np.ones(x.shape[0])
        if score_sample_weights is None:
            score_sample_weights = np.ones(x.shape[0])

        cv_generator = KFold(n_splits=n_splits)
        baseline_scores, shuffled_scores = pd.Series(dtype=float), pd.DataFrame(columns=clusters.keys())

        for i, (train, test) in enumerate(cv_generator.split(X=x)):
            print(f"Fold {i} start ...")

            x_train, y_train, train_weights = x.iloc[train, :], y.iloc[train], train_sample_weights[train]
            x_test, y_test, test_weights = x.iloc[test, :], y.iloc[test], score_sample_weights[test]

            classifier_fit = classifier.fit(X=x_train, y=y_train, sample_weight=train_weights)
            prediction_probability = classifier_fit.predict_proba(x_test)

            baseline_scores[i] = -log_loss(y_test, prediction_probability, labels=classifier.classes_, sample_weight=test_weights)

            for cluster_name in shuffled_scores.columns:
                x_test_shuffled = x_test.copy(deep=True)
                for feature in clusters[cluster_name]:
                    np.random.shuffle(x_test_shuffled[feature].values)
                prob = classifier_fit.predict_proba(x_test_shuffled)
                shuffled_scores.loc[i, cluster_name] = -log_loss(y_test, prob, labels=classifier.classes_)

        importances = (-1 * shuffled_scores).add(baseline_scores, axis=0)
        importances /= -1 * shuffled_scores

        # Central Limit Theorem for standard deviation
        importances = pd.concat(
            {"Mean": importances.mean(), "StandardDeviation": importances.std() * (importances.shape[0] ** -0.5)}, axis=1
        )

        importances.index = ["C_" + str(i) for i in importances.index]
        return importances



from RiskLabAI.data.structures.abstract_information_driven_bars import *
from RiskLabAI.utils.constants import *


class AbstractImbalanceBars(AbstractInformationDrivenBars):
    """
    Abstract class that contains the imbalance properties which are shared between the subtypes.
    This class subtypes are as follows:
        1- ExpectedImbalanceBars
        2- FixedImbalanceBars

    The class implements imbalance bars sampling logic as explained on page 29,30 of Advances in Financial Machine Learning.
    """

    def __init__(
            self,
            bar_type: str,
            window_size_for_expected_n_ticks_estimation: int,
            window_size_for_expected_imbalance_estimation: int,
            initial_estimate_of_expected_n_ticks_in_bar: int,
            analyse_thresholds: bool
    ):
        """
        AbstractImbalanceBars constructor function
        :param bar_type: type of bar. e.g. expected_dollar_imbalance_bars, fixed_tick_imbalance_bars etc.
        :param window_size_for_expected_n_ticks_estimation: window size used to estimate number of ticks expectation
        :param initial_estimate_of_expected_n_ticks_in_bar: initial estimate of number of ticks expectation window size
        :param window_size_for_expected_imbalance_estimation: window size used to estimate imbalance expectation
        :param analyse_thresholds: whether return thresholds values (θ, number of ticks expectation, imbalance expectation) in a tabular format
        """

        super().__init__(
            bar_type,
            window_size_for_expected_n_ticks_estimation,
            initial_estimate_of_expected_n_ticks_in_bar,
            window_size_for_expected_imbalance_estimation
        )

        self.imbalance_bars_statistics = {
            CUMULATIVE_θ: 0,
            EXPECTED_IMBALANCE: np.nan,

            # previous bars number of ticks list and previous tick imbalances list
            PREVIOUS_BARS_N_TICKS_LIST: [],
            PREVIOUS_TICK_IMBALANCES_LIST: [],
        }

        if analyse_thresholds:
            self.analyse_thresholds = []
        else:
            self.analyse_thresholds = None

    def construct_bars_from_data(self, data: Union[list, tuple, np.ndarray]) -> list:
        """
        The function is used to construct bars from input ticks data.
        :param data: tabular data that contains date_time, price, and volume columns
        :return: constructed bars
        """

        bars_list = []
        for tick_data in data:
            self.tick_counter += 1

            date_time, price, volume = tuple(tick_data)
            tick_rule = self._tick_rule(price)
            self.update_base_fields(price, tick_rule, volume)

            # imbalance calculations
            imbalance = self._imbalance_at_tick(price, tick_rule, volume)
            self.imbalance_bars_statistics[PREVIOUS_TICK_IMBALANCES_LIST].append(imbalance)
            self.imbalance_bars_statistics[CUMULATIVE_θ] += imbalance

            # initialize expected imbalance first time, where expected imbalance estimation is nan
            if np.isnan(self.imbalance_bars_statistics[EXPECTED_IMBALANCE]):
                self.imbalance_bars_statistics[EXPECTED_IMBALANCE] = self._ewma_expected_imbalance(
                    self.imbalance_bars_statistics[PREVIOUS_TICK_IMBALANCES_LIST],
                    self.information_driven_bars_statistics[EXPECTED_IMBALANCE_WINDOW],
                    warm_up=True,  # this field just apply for run bars
                )

            if self.analyse_thresholds is not None:
                self.imbalance_bars_statistics['timestamp'] = date_time
                self.analyse_thresholds.append(dict(self.imbalance_bars_statistics))

            # is construction condition met to construct next bar or not
            expected_ticks_number = self.information_driven_bars_statistics[EXPECTED_TICKS_NUMBER]
            expected_imbalance = self.imbalance_bars_statistics[EXPECTED_IMBALANCE]
            threshold = expected_ticks_number * np.abs(expected_imbalance)
            is_construction_condition_met = self._bar_construction_condition(threshold)

            if is_construction_condition_met:
                next_bar = self._construct_next_bar(
                    date_time,
                    self.tick_counter,
                    price,
                    self.high_price,
                    self.low_price,
                    threshold
                )

                bars_list.append(next_bar)

                self.imbalance_bars_statistics[PREVIOUS_BARS_N_TICKS_LIST].append(
                    self.base_statistics[CUMULATIVE_TICKS])

                # update expected number of ticks based on formed bars
                self.information_driven_bars_statistics[EXPECTED_TICKS_NUMBER] = self._expected_number_of_ticks()

                # update expected imbalance (ewma)
                self.imbalance_bars_statistics[EXPECTED_IMBALANCE] = self._ewma_expected_imbalance(
                    self.imbalance_bars_statistics[PREVIOUS_TICK_IMBALANCES_LIST],
                    self.information_driven_bars_statistics[EXPECTED_IMBALANCE_WINDOW],
                    warm_up=True,  # this field just apply for run bars
                )

                # reset cached fields
                self._reset_cached_fields()

        return bars_list

    def _bar_construction_condition(
            self,
            threshold
    ):
        """
        Compute the condition of whether next bar should sample with current and previous tick datas or not.
        :return: whether next bar should form with current and previous tick datas or not.
        """

        if not np.isnan(self.imbalance_bars_statistics[EXPECTED_IMBALANCE]):
            cumulative_θ = self.imbalance_bars_statistics[CUMULATIVE_θ]
            condition_is_met = np.abs(cumulative_θ) > threshold

            return condition_is_met

        else:
            return False

    def _reset_cached_fields(
            self
    ):
        """
        This function are used (directly or override) by all concrete or abstract subtypes. The function is used to reset cached fields in bars construction process when next bar is sampled.
        :return:
        """
        super()._reset_cached_fields()
        self.imbalance_bars_statistics[CUMULATIVE_θ] = 0

    @abstractmethod
    def _expected_number_of_ticks(self) -> Union[float, int]:
        """
        Calculate number of ticks expectation when new imbalance bar is sampled.

        :return: number of ticks expectation.
        """



import numpy as np
from typing import Optional

def pca_weights(
    cov: np.ndarray,
    risk_distribution: Optional[np.ndarray] = None,
    risk_target: float = 1.0
) -> np.ndarray:
    """
    Calculates hedging weights using covariance, risk distribution, and risk target.

    The function uses Principal Component Analysis (PCA) to determine the weights.
    If the risk distribution is not provided, all risk is allocated to the principal
    component with the smallest eigenvalue.

    :param cov: Covariance matrix.
    :param risk_distribution: Risk distribution, defaults to None.
    :param risk_target: Risk target, defaults to 1.0.

    :return: Weights calculated based on PCA.

    .. note::
       Reference:
       De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons.
       Methodology 36.

    .. math::
       w = EV . \sqrt{\frac{\rho T}{\lambda}}

       where:
       - :math:`w` are the weights.
       - :math:`EV` is the matrix of eigenvectors.
       - :math:`\rho` is the risk distribution.
       - :math:`T` is the risk target.
       - :math:`\lambda` is the eigenvalues.
    """

    # Calculate the eigenvalues and eigenvectors of the covariance matrix
    eigen_values, eigen_vectors = np.linalg.eigh(cov)
    # Sort eigenvalues in descending order
    indices = eigen_values.argsort()[::-1]
    eigen_values, eigen_vectors = eigen_values[indices], eigen_vectors[:, indices]

    # If risk_distribution is not provided, allocate all risk to the principal component
    # with the smallest eigenvalue.
    if risk_distribution is None:
        risk_distribution = np.zeros(cov.shape[0])
        risk_distribution[-1] = 1.0

    # Compute loads (allocation in the new orthogonal basis)
    loads = risk_target * (risk_distribution / eigen_values) ** 0.5
    # Calculate weights
    weights = np.dot(eigen_vectors, np.reshape(loads, (-1, 1)))

    return weights.flatten()



import numpy as np
import pandas as pd
import statsmodels.api as sm
import sklearn.datasets as Datasets


def generate_ar_time_series(n, ar_params, ma_params, sigma=1.0):
    """
    Generate an autoregressive (AR) time series.

    Parameters:
    - n: Length of the time series
    - ar_params: List of AR parameters [phi1, phi2, ..., phi_p]
    - sigma: Standard deviation of the white noise

    Returns:
    - time_series: Generated AR time series
    """
    # Create the AR coefficients for the model
    ar = np.r_[1, -np.array(ar_params)]
    ma = np.r_[1, np.array(ma_params)]

    # Generate the ARMA process
    arma_process = sm.tsa.ArmaProcess(ar, ma)
    time_series = arma_process.generate_sample(nsample=n, scale=sigma)

    return time_series


def create_lagged_data(time_series, k):
    """
    Create a DataFrame with the time series and its k lags.

    Parameters:
    - time_series: The generated time series
    - k: Number of lags

    Returns:
    - df: DataFrame containing the time series and its lags
    """
    data = {'y': time_series}
    for i in range(1, k + 1):
        data[f'lag_{i}'] = np.roll(time_series, i)

    df = pd.DataFrame(data)
    df = df.iloc[k:]  # Drop the initial rows with NaN values due to the roll

    return df.to_numpy()

def generate_cross_sectional_dataset(
    n_informative: int = 5, 
    n_redundant: int = 25,  
    n_noise: int = 5, 
    n_samples: int = 10000,  
    random_state: int = 41,  
    sigma_std: float = 0.1, 
    n_clusters_per_class: int = 2,
) -> tuple:
    n_features = n_informative + n_redundant + n_noise
    np.random.seed(random_state)
    X, y = Datasets.make_classification(
        n_samples=n_samples,
        n_features=n_features-n_redundant,
        n_informative=n_informative,
        n_redundant=0,
        shuffle=False,
        random_state=random_state,
        n_clusters_per_class=n_clusters_per_class,
    )

    columns = [f"I{i}" for i in range(n_informative)]
    columns += [f"N{i}" for i in range(n_noise)]
    X, y = pd.DataFrame(X, columns=columns), pd.Series(y)
    i = np.random.choice(range(n_informative), size=n_redundant)
    for k, j in enumerate(i):
        X[f"R{k} (from I{j})"] = X[f"I{j}"] + np.random.normal(size=X.shape[0]) * sigma_std

    return X, y

def generate_multivariate_time_series_dataset(
    n_informative: int = 10,  # number of informative features
    n_redundant: int = 10,  # number of redundant features
    n_noise: int = 20,
    n_samples: int = 10000,  # number of sample to generate
    n_time_steps: int = 2, # Two business years
    random_state: int = 41,  # random state
    sigma_std: float = 0.1,  # standard deviation of generation
    time_series_params = [[0.3], [0.3], 0.1], # AR parameter (phi1)
):
    n_lag = n_time_steps - 1

    informative_features = np.zeros(shape=(n_samples, n_informative, n_time_steps))
    for k in range(n_informative):
        time_series = generate_ar_time_series(n_samples + n_lag, time_series_params[0], time_series_params[1], 1)
        informative_features[:, k, :] = create_lagged_data(time_series, n_lag)

    np.random.seed(random_state)
    i = np.random.choice(range(n_informative), size=n_redundant)
    np.random.seed()
    
    linear_redundant_features = np.zeros((n_samples, n_redundant, n_time_steps))
    linear_redundant_features_from_ = []

    for k, j in enumerate(i):
        linear_redundant_features[:, k, :] = informative_features[:, j, :] + np.random.normal(size=(n_samples, n_time_steps)) * sigma_std
        linear_redundant_features_from_.append(j)

    # Generate noise features
    noise_features = np.zeros(shape=(n_samples, n_noise, n_time_steps))
    for k in range(n_noise):
        time_series = generate_ar_time_series(n_samples + n_lag, 0.5, 0.5, 1)
        noise_features[:, k, :] = create_lagged_data(time_series, n_lag)

    # Concatenate all features
    X = np.concatenate([
        informative_features,
        linear_redundant_features,
        noise_features],
        axis=1)

    columns = [f"I{i} lag{j}" for i in range(n_informative) for j in range(n_time_steps)]
    columns += [f"R{i} from I{k} lag{j}" for i, k in enumerate(linear_redundant_features_from_) for j in range(n_time_steps)]
    columns += [f"N{i} lag{j}" for i in range(n_noise) for j in range(n_time_steps)]

    X = X.reshape((n_samples, -1))
    X = pd.DataFrame(X, columns=columns)

    I = X.iloc[:, :n_informative * n_time_steps]
    I = (I - I.mean()) / I.std()
    weights = np.ones(shape=len(I.columns))

    # Calculate linear combination of features plus some noise
    linear_combination = I @ weights
    # Generate binary target variable based on probabilities
    y = (linear_combination > 0).astype(int)

    y = pd.Series(y)

    return X, y




import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from typing import Dict, Tuple, Generator, Any, Optional
from copy import deepcopy
from joblib import Parallel, delayed
import warnings
from sklearn.exceptions import ConvergenceWarning
from .kfold import KFold

class WalkForward(KFold):
    """
    WalkForward Cross-Validator for Time Series Data.

    This cross-validator provides train/test indices meant to split time series data 
    in a "walk-forward" manner, which is suitable for time series forecasting tasks. 
    In each split, the training set progressively grows in size (subject to the optional
    maximum size constraint) while the test set remains roughly constant in size. 
    A gap can be optionally introduced between the training and test set to simulate 
    forecasting on unseen future data after a certain interval.

    The WalkForward cross-validator is inherently different from traditional K-Fold
    cross-validation which shuffles and splits the dataset into train/test without 
    considering the time order. In time series tasks, ensuring that the model is 
    trained on past data and validated on future data is crucial. This cross-validator 
    achieves that by progressively walking forward in time through the dataset.
    """

    def __init__(
        self,
        n_splits: int = 5,
        max_train_size: int = None,
        gap: int = 0
    ) -> None:
        """
        Initialize the TimeSeriesWalkForward cross-validator.

        Parameters:
        -----------
        n_splits : int, default=5
            Number of splits/folds. Must be at least 2.
        
        max_train_size : int, optional
            Maximum number of observations allowed in the training dataset.
            If provided, the most recent `max_train_size` observations are used 
            for training.

        gap : int, default=0
            Number of observations to skip between the end of the training data 
            and the start of the test data. Useful for simulating forecasting 
            scenarios where the test data is not immediately after the training data.
        """
        super().__init__(n_splits=n_splits)  # Initializing the parent class with n_splits
        self.max_train_size = max_train_size
        self.gap = gap


    def _single_split(
        self,
        single_data: pd.DataFrame,
    ) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
        """
        Splits a single data set into train-test indices.

        This function provides train-test indices to split the data into train/test sets
        by respecting the time order (if applicable) and the specified number of splits.

        :param single_data: Input dataset.
        :type single_data: pd.DataFrame

        :return: Generator yielding train-test indices.
        :rtype: Generator[Tuple[np.ndarray, np.ndarray], None, None]
        """

        indices = np.arange(single_data.shape[0])
        gap = self.gap
        max_train_size = self.max_train_size

        for test_indices in np.array_split(indices, self.n_splits):
            # Ensure that train indices come before test indices
            last_test_index = test_indices[-1]
            train_indices = np.setdiff1d(indices, test_indices)
            train_indices = train_indices[train_indices < last_test_index]
            
            # Exclude gap from the end of train set
            if self.gap:
                train_indices = train_indices[:-gap]
            
            # Ensuring the training set does not exceed max_train_size
            if self.max_train_size and len(train_indices) > max_train_size:
                train_indices = train_indices[-max_train_size:]
            
            yield train_indices, test_indices    


    def _single_backtest_predictions(
        self,
        single_estimator: Any,
        single_data: pd.DataFrame,
        single_labels: pd.Series,
        single_weights: Optional[np.ndarray] = None,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Dict[str, np.ndarray]:
        """
        Obtain predictions for a single dataset during backtesting.

        This function leverages parallel computation to train and predict on different train-test splits
        of a single dataset using a given estimator. It utilizes the `_single_split` method to generate
        the train-test splits.

        :param single_estimator: Estimator or model to be trained and used for predictions.
        :type single_estimator: Any
        :param single_data: Data of the single dataset.
        :type single_data: pd.DataFrame
        :param single_labels: Labels corresponding to the single dataset.
        :type single_labels: pd.Series
        :param single_weights: Weights for the observations in the single dataset.
                            Defaults to equally weighted if not provided.
        :type single_weights: np.ndarray, optional
        :param predict_probability: If True, predict the probability of forecasts.
        :type predict_probability: bool
        :param n_jobs: The number of jobs to run in parallel. Default is 1.
        :type n_jobs: int, optional
        :return: Predictions structured in a dictionary for the backtest paths.
        :rtype: Dict[str, np.ndarray]
        """

        if single_weights is None:
            single_weights = np.ones(len(single_data))

        def train_test_single_estimator(
            single_estimator_: Any,
            train_indices: np.ndarray,
            test_indices: np.ndarray
        ) -> np.ndarray:
            
            if len(train_indices) == 0:
                if predict_probability:
                    n_columns = len(np.unique(single_labels)) if predict_probability else 1
                    return np.full((len(test_indices), n_columns), np.nan)
        
                else:
                    return np.full(len(test_indices), np.nan)
            
            X_train = single_data.iloc[train_indices]
            y_train = single_labels.iloc[train_indices]
            weights_train = single_weights[train_indices]

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=ConvergenceWarning)
                try:
                    single_estimator_.fit(X_train, y_train, sample_weight=weights_train)

                except TypeError:
                    single_estimator_.fit(X_train, y_train)

            X_test = single_data.iloc[test_indices]

            if predict_probability:
                return single_estimator_.predict_proba(X_test)

            else:
                return single_estimator_.predict(X_test)

        path_data = Parallel(n_jobs=n_jobs)(
            delayed(train_test_single_estimator)(
                deepcopy(single_estimator), train_indices, test_indices
            ) for train_indices, test_indices in self._single_split(single_data)
        )

        paths_predictions = {'Path 1': np.concatenate(path_data)}

        return paths_predictions        



import pandas as pd
import numpy as np


def compute_eigenvectors(
        dot_product: np.ndarray,
        explained_variance_threshold: float
) -> pd.DataFrame:
    """
    Compute eigenvalues and eigenvectors for orthogonal features.

    :param dot_product: Input dot product matrix.
    :type dot_product: np.ndarray
    :param explained_variance_threshold: Threshold for cumulative explained variance.
    :type explained_variance_threshold: float
    :return: DataFrame containing eigenvalues, eigenvectors, and cumulative explained variance.
    :rtype: pd.DataFrame
    """
    eigenvalues, eigenvectors = np.linalg.eigh(dot_product)

    eigen_dataframe = pd.DataFrame({
        "Index": [f"PC {i}" for i in range(1, len(eigenvalues) + 1)],
        "EigenValue": eigenvalues,
        "EigenVector": [ev for ev in eigenvectors]
    })

    eigen_dataframe = eigen_dataframe.sort_values("EigenValue", ascending=False)

    cumulative_variance = np.cumsum(eigen_dataframe["EigenValue"]) / np.sum(eigen_dataframe["EigenValue"])
    eigen_dataframe["CumulativeVariance"] = cumulative_variance

    index = cumulative_variance.searchsorted(explained_variance_threshold)

    eigen_dataframe = eigen_dataframe.iloc[:index + 1, :]

    return eigen_dataframe


def orthogonal_features(
        features: np.ndarray,
        variance_threshold: float = 0.95
) -> tuple:
    """
    Compute orthogonal features using eigenvalues and eigenvectors.

    :param features: Features matrix.
    :type features: np.ndarray
    :param variance_threshold: Threshold for cumulative explained variance, default is 0.95.
    :type variance_threshold: float
    :return: Tuple containing orthogonal features and eigenvalues information.
    :rtype: tuple
    """
    normalized_features = (features - features.mean(axis=0)) / features.std(axis=0)
    dot_product = normalized_features.T @ normalized_features
    eigen_dataframe = compute_eigenvectors(dot_product, variance_threshold)

    transformation_matrix = np.vstack(eigen_dataframe["EigenVector"].values).T

    orthogonal_features = normalized_features @ transformation_matrix

    return orthogonal_features, eigen_dataframe



# from feature_importance_strategy import FeatureImportanceStrategy
from RiskLabAI.features.feature_importance.feature_importance_strategy import FeatureImportanceStrategy
import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestClassifier
from typing import Dict, List


class ClusteredFeatureImportanceMDI(FeatureImportanceStrategy):

    def __init__(
            self,
            classifier: RandomForestClassifier,
            clusters: Dict[str, List[str]],
            x: pd.DataFrame,
            y: pd.Series
    ):
        """
        Initialize the ClusteredFeatureImportanceMDI class.

        :param classifier: The Random Forest classifier.
        :param clusters: A dictionary where the keys are the cluster names 
                         and the values are lists of features in each cluster.
        :param x: The features DataFrame.
        :param y: The target Series.
        """
        self.classifier = classifier
        self.clusters = clusters
        classifier.fit(x, y)

    def group_mean_std(
            self,
            dataframe: pd.DataFrame,
            clusters: Dict[str, List[str]]
    ) -> pd.DataFrame:
        """
        Calculate the mean and standard deviation for clusters.

        :param dataframe: A DataFrame of importances.
        :param clusters: A dictionary of cluster definitions.
        
        :return: A DataFrame with mean and standard deviation for each cluster.

        Using Central Limit Theorem for standard deviation:

        .. math::

            \\text{{StandardDeviation}} = \\text{{std}} \\times n^{-0.5}

        """
        output = pd.DataFrame(columns=["Mean", "StandardDeviation"])

        for cluster_name, feature_names in clusters.items():
            cluster_data = dataframe[feature_names].sum(axis=1)
            output.loc["C_" + str(cluster_name), "Mean"] = cluster_data.mean()
            output.loc["C_" + str(cluster_name), "StandardDeviation"] = (
                cluster_data.std() * (cluster_data.shape[0] ** -0.5)
            )

        return output

    def compute(self) -> pd.DataFrame:
        """
        Compute aggregated feature importances for clusters.

        :return: A DataFrame with aggregated importances for clusters.
        """
        importances_dict = {i: tree.feature_importances_ for i, tree in enumerate(self.classifier.estimators_)}
        importances_df = pd.DataFrame.from_dict(importances_dict, orient="index")
        importances_df.columns = self.classifier.feature_names_in_
        importances_df.replace(0, np.nan, inplace=True)  # because max_features=1

        aggregated_importances = self.group_mean_std(importances_df, self.clusters)
        aggregated_importances /= aggregated_importances["Mean"].sum()

        return aggregated_importances



from typing import Generator, Optional, Tuple, Union, Dict, List, Any
from itertools import combinations
import pandas as pd
import numpy as np
from collections import ChainMap
from joblib import Parallel, delayed
import warnings
from sklearn.exceptions import ConvergenceWarning
from joblib_progress import joblib_progress
from sklearn.base import clone

from .combinatorial_purged import CombinatorialPurged

class AdaptiveCombinatorialPurged(CombinatorialPurged):
    def __init__(
        self,
        n_splits: int,
        n_test_groups: int,
        times: Union[pd.Series, Dict[str, pd.Series]],
        embargo: float = 0,
        n_subsplits: int = 3,
        external_feature: Union[pd.Series, Dict[str, pd.Series]] = None,
        lower_quantile: float = 0.25,
        upper_quantile: float = 0.75,
        subtract_border_adjustments: bool = True
    ):
        """
        Initialize the AdaptiveCombinatorialPurged class.

        Parameters
        ----------
        n_splits : int
            Number of splits/groups to partition the data into.
        n_test_groups : int
            Size of the testing set in terms of groups.
        times : Union[pd.Series, Dict[str, pd.Series]]
            The timestamp series associated with the labels.
        embargo : float
            The embargo rate for purging.
        n_subsplits : int
            Number of subsplits within each split segment.
        external_feature : Union[pd.Series, Dict[str, pd.Series]]
            The external feature based on which the adaptive splitting is performed.
        lower_quantile : float
            The lower quantile threshold for adjusting the split segments.
        upper_quantile : float
            The upper quantile threshold for adjusting the split segments.
        subtract_border_adjustments : bool
            Flag to determine whether to subtract border adjustments instead of adding.
        """
        super().__init__(n_splits, n_test_groups, times, embargo)
        self.n_subsplits = n_subsplits
        self.external_feature = external_feature
        self.lower_quantile = lower_quantile
        self.upper_quantile = upper_quantile
        self.subtract_border_adjustments = subtract_border_adjustments


    def _validate_input(
        self,
        single_times: pd.Series,
        single_data: pd.DataFrame,
        single_external_feature: pd.Series
    ) -> None:
        """
        Validate that the input data, times, and external feature share the same index.

        This function checks if the provided data, times, and external feature have the same index.
        If they do not match, it raises a `ValueError`.

        :param single_times: Time series data to be validated.
        :type single_times: pd.Series
        :param single_data: Dataset with which the times should align.
        :type single_data: pd.DataFrame
        :param single_external_feature: External feature series to be validated.
        :type single_external_feature: pd.Series
        :raises ValueError: If the indices of the data, times, and external feature do not match.
        :return: None
        """
        if not single_data.index.equals(single_times.index) or not single_data.index.equals(single_external_feature.index):
            raise ValueError('Data, through date values, and external feature must have the same index')
        

    def _single_adaptive_split_segments(
        self, 
        indices: np.ndarray, 
        single_external_feature: pd.Series
    ) -> List[np.ndarray]:
        """
        Adaptively split data indices based on the external feature's values and quantile thresholds.

        Parameters
        ----------
        indices : np.ndarray
            Array of data indices to be split.
        single_external_feature : pd.Series
            The external feature based on which the adaptive splitting is performed.

        Returns
        -------
        split_segments : List[np.ndarray]
            List of adaptively split data indices.
        """
        external_values = single_external_feature.iloc[indices].values

        # Calculate the quantile thresholds
        lower_threshold = np.percentile(external_values, self.lower_quantile * 100)
        upper_threshold = np.percentile(external_values, self.upper_quantile * 100)

        # Partition each group into n_subsplits
        subsplits = np.array_split(indices, self.n_splits * self.n_subsplits)

        # Get the starting points of each subsplit
        subsplit_starts = np.array([split[0] for split in subsplits])

        # Determine the initial borders between subsplits
        borders = np.arange(self.n_subsplits, len(subsplit_starts), self.n_subsplits)

        # Vectorized comparison of external feature values at borders
        border_values = external_values[subsplit_starts[borders]]

        # Determine adjustments based on quantiles
        border_adjustments = np.zeros_like(borders)
        border_adjustments[border_values < lower_threshold] = -1
        border_adjustments[border_values > upper_threshold] = 1

        # Adjust the borders based on the subtract_border_adjustments flag
        if self.subtract_border_adjustments:
            adjusted_borders = borders - border_adjustments
        else:
            adjusted_borders = borders + border_adjustments

        # Ensure the borders are within valid range
        adjusted_borders = np.clip(adjusted_borders, 1, len(subsplit_starts) - 1)

        # Create the final split segments
        split_segments = np.split(indices, adjusted_borders)

        return split_segments
    

    def _single_split(
        self,
        single_times: pd.Series,
        single_data: pd.DataFrame,
        single_external_feature: pd.Series
    ) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
        """
        Splits data into train and test indices based on the defined combinatorial splits.

        This function is used to generate multiple train-test splits based on the combinatorial
        cross-validation method. It ensures that each train-test split is properly purged and
        embargoed to prevent data leakage.

        :param single_times: Timestamp series associated with the labels.
        :param single_data: The input data to be split.
        :param single_external_feature: External feature series used for adaptive splitting.

        :return: Generator that yields tuples of (train indices, test indices).

        .. note:: The function validates the input, and uses combinatorial cross-validation method to
                produce the train-test splits.
        """
        self._validate_input(single_times, single_data, single_external_feature)

        indices = np.arange(single_data.shape[0])
        split_segments = self._single_adaptive_split_segments(indices, single_external_feature)
        combinations_ = list(combinations(range(self.n_splits), self.n_test_groups))

        all_combinatorial_splits = list(
            CombinatorialPurged._combinatorial_splits(combinations_, split_segments)
        )

        for test_indices in all_combinatorial_splits:
            train_indices = self._get_train_indices(test_indices, single_times)
            yield train_indices, test_indices


    def split(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        labels: Optional[Union[pd.Series, Dict[str, pd.Series]]] = None,
        groups: Optional[np.ndarray] = None
    ) -> Union[
        Generator[Tuple[np.ndarray, np.ndarray], None, None],
        Generator[Tuple[str, Tuple[np.ndarray, np.ndarray]], None, None]
    ]:
        """
        Split multiple datasets into train and test sets.

        This function either splits a single dataset or multiple datasets considering
        purging and embargo.

        :param data: Dataset or dictionary of datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
        :param labels: Labels corresponding to the datasets, if available.
        :type labels: Optional[Union[pd.Series, Dict[str, pd.Series]]]
        :param groups: Group information, if available.
        :type groups: Optional[np.ndarray]
        :return: Train and test indices or key with train and test indices for multiple datasets.
        :rtype: Union[Generator[Tuple[np.ndarray, np.ndarray], None, None],
                    Generator[Tuple[str, Tuple[np.ndarray, np.ndarray]], None, None]]
        """
        if self.is_multiple_datasets:
            for key in self.times:
                for train_indices, test_indices in self._single_split(
                    self.times[key], data[key], self.external_feature[key]
                ):
                    yield key, (train_indices, test_indices)
        else:
            for train_indices, test_indices in self._single_split(
                self.times, data, self.external_feature
            ):
                yield train_indices, test_indices        


    def _combinations_and_path_locations_and_split_segments(
        self,
        data: pd.DataFrame,
        single_external_feature: pd.Series
    ) -> Tuple[List[Tuple[int, ...]], Dict[str, np.ndarray], List[np.ndarray]]:
        """
        Generate combinations, path locations, and split segments for the data.

        This function is a helper that computes necessary components for combinatorial cross-validation.

        :param data: The input dataframe to generate combinations, path locations, and split segments.
        :param single_external_feature: External feature series used for adaptive splitting.

        :return: Tuple containing combinations, path locations, and split segments.

        .. math::
        \\text{combinations} = \\binom{n}{k}
        """
        combinations_ = list(combinations(range(self.n_splits), self.n_test_groups))
        locations = CombinatorialPurged._path_locations(self.n_splits, combinations_)

        indices = np.arange(data.shape[0])
        split_segments = self._single_adaptive_split_segments(indices, single_external_feature)

        return combinations_, locations, split_segments

    def _single_backtest_paths(
        self,
        single_times: pd.Series,
        single_data: pd.DataFrame,
        single_external_feature: pd.Series
    ) -> Dict[str, List[Dict[str, List[np.ndarray]]]]:
        """
        Generate the backtest paths for given input data.

        This function creates multiple backtest paths based on combinatorial splits, where
        each path represents a sequence of train-test splits. It ensures that data leakage
        is prevented by purging and applying embargo to the train-test splits.

        :param single_times: Timestamp series associated with the data.
        :param single_data: Input data on which the backtest paths are based.
        :param single_external_feature: External feature series used for adaptive splitting.

        :return: A dictionary where each key is a backtest path name, and the value is
                a list of dictionaries with train and test index arrays.

        .. note:: This function relies on combinatorial cross-validation for backtesting to
                generate multiple paths of train-test splits.
        """
        self._validate_input(single_times, single_data, single_external_feature)

        paths = {}
        combinations_, locations, split_segments = self._combinations_and_path_locations_and_split_segments(single_data, single_external_feature)
        all_combinatorial_splits = list(
            CombinatorialPurged._combinatorial_splits(combinations_, split_segments)
        )

        for path_num, locs in locations.items():
            path_data = []

            for (G, S) in locs:
                # Use all_combinatorial_splits[S] to determine potential training indices
                # Now, we derive the actual training set after purge and embargo
                train_indices = self._get_train_indices(all_combinatorial_splits[S], single_times)

                path_data.append({
                    "Train": np.array(train_indices),
                    "Test": split_segments[G],
                })

            paths[f"Path {path_num}"] = np.array(path_data)

        return paths
    

    def backtest_paths(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
    ) -> Union[
        Dict[str, List[Dict[str, np.ndarray]]],
        Dict[str, Dict[str, List[Dict[str, List[np.ndarray]]]]]
    ]:
        """
        Generate backtest paths for single or multiple datasets.

        This function checks whether multiple datasets are being used. If so, it iterates through each
        dataset, generating backtest paths using the `_single_backtest_paths` method. Otherwise, it directly
        returns the backtest paths for the single dataset.

        :param data: Input data on which the backtest paths are based.
                    Can be either a single DataFrame or a dictionary of DataFrames for multiple datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]

        :return: A dictionary where each key is a backtest path name, and the value is
                a list of dictionaries with train and test index arrays. For multiple datasets,
                a nested dictionary structure is returned.
        :rtype: Union[
            Dict[str, List[Dict[str, np.ndarray]]],
            Dict[str, Dict[str, List[Dict[str, List[np.ndarray]]]]]
        ]

        """
        if self.is_multiple_datasets:
            multiple_paths = {}
            for key in self.times:
                multiple_paths[key] = self._single_backtest_paths(
                    self.times[key], data[key], self.external_feature[key]
                )

            return multiple_paths
        else:
            return self._single_backtest_paths(self.times, data, self.external_feature)


    def _single_backtest_predictions(
        self,
        single_estimator: Any,
        single_times: pd.Series,
        single_data: pd.DataFrame,
        single_labels: pd.Series,
        single_weights: np.ndarray,
        single_external_feature: pd.Series,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Dict[str, np.ndarray]:
        """
        Generate predictions for a single backtest using combinatorial splits.

        This method calculates predictions across various paths created by combinatorial splits
        of the data. For each combinatorial split, a separate estimator is trained and then used
        to predict on the corresponding test set.

        :param single_estimator: The machine learning model or estimator to be trained.
        :param single_times: Timestamps corresponding to the data points.
        :param single_data: Input data on which the model is trained and predictions are made.
        :param single_labels: Labels corresponding to the input data.
        :param single_weights: Weights for each data point.
        :param single_external_feature: External feature series used for adaptive splitting.
        :param predict_probability: If True, predict the probability of forecasts.
        :type predict_probability: bool
        :param n_jobs: Number of CPU cores to use for parallelization. Default is 1.

        :return: A dictionary where keys are path names and values are arrays of predictions.

        .. note:: This function relies on internal methods (e.g., `_get_train_indices`)
                to manage data splits and training.

        .. note:: Parallelization is used to speed up the training of models for different splits.
        """
        self._validate_input(single_times, single_data, single_external_feature)

        if single_weights is None:
            single_weights = pd.Series(np.ones((len(single_data),)))

        paths_predictions = {}
        combinations_, locations, split_segments = self._combinations_and_path_locations_and_split_segments(single_data, single_external_feature)

        def train_single_estimator(
            single_estimator_,
            combinatorial_test_indices
        ):
            combinatorial_train_indices = self._get_train_indices(combinatorial_test_indices, single_times)

            X_train = single_data.iloc[combinatorial_train_indices]
            y_train = single_labels.iloc[combinatorial_train_indices]
            weights_train = single_weights.iloc[combinatorial_train_indices]

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=ConvergenceWarning)
                try:
                    single_estimator_.fit(X_train, y_train, sample_weight=weights_train)

                except (TypeError, ValueError):
                    single_estimator_.fit(X_train, y_train)

            return single_estimator_

        # with joblib_progress("Backtesting...", total=self.get_n_splits()): 
        combinatorial_trained_estimators = Parallel(n_jobs=n_jobs)(
            delayed(train_single_estimator)(clone(single_estimator), combinatorial_test_indices)
            for combinatorial_test_indices in CombinatorialPurged._combinatorial_splits(combinations_, split_segments))


        def get_path_data(
            path_num,
            locs
        ):
            path_data = []

            for (G, S) in locs:
                test_indices = split_segments[G]
                X_test = single_data.iloc[test_indices]

                if predict_probability:
                    predictions = combinatorial_trained_estimators[S].predict_proba(X_test)

                else:
                    predictions = combinatorial_trained_estimators[S].predict(X_test)

                path_data.extend(predictions)

            path_data = {f"Path {path_num}" : np.array(path_data)}

            return path_data

        paths_predictions = Parallel(n_jobs=n_jobs)(delayed(get_path_data)(path_num, locs) for path_num, locs in locations.items())
        paths_predictions = dict(ChainMap(*reversed(paths_predictions)))

        return paths_predictions
        

    def backtest_predictions(
        self,
        estimator: Union[Any, Dict[str, Any]],
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        labels: Union[pd.Series, Dict[str, pd.Series]],
        sample_weights: Optional[Union[np.ndarray, Dict[str, np.ndarray]]] = None,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Union[Dict[str, np.ndarray], Dict[str, Dict[str, np.ndarray]]]:
        """
        Generate backtest predictions for single or multiple datasets.

        For each dataset, this function leverages the `_single_backtest_predictions` method to obtain
        predictions for different train-test splits using the given estimator.

        :param estimator: Model or estimator to be trained and used for predictions.
                        Can be a single estimator or a dictionary of estimators for multiple datasets.
        :type estimator: Union[Any, Dict[str, Any]]
        :param data: Input data for training and testing. Can be a single dataset or
                    a dictionary of datasets for multiple datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
        :param labels: Target labels for training and testing. Can be a single series or
                    a dictionary of series for multiple datasets.
        :type labels: Union[pd.Series, Dict[str, pd.Series]]
        :param sample_weights: Weights for the observations in the dataset(s).
                            Can be a single array or a dictionary of arrays for multiple datasets.
                            Defaults to None, which means equal weights for all observations.
        :type sample_weights: Optional[Union[np.ndarray, Dict[str, np.ndarray]]]
        :param predict_probability: If True, predict the probability of forecasts.
        :type predict_probability: bool
        :param n_jobs: The number of jobs to run in parallel. Default is 1.
        :type n_jobs: int, optional
        :return: Backtest predictions structured in a dictionary (or nested dictionaries for multiple datasets).
        :rtype: Union[Dict[str, np.ndarray], Dict[str, Dict[str, np.ndarray]]]
        """
        if self.is_multiple_datasets:
            multiple_paths_predictions = {}
            for key in self.times:
                multiple_paths_predictions[key] = self._single_backtest_predictions(
                    estimator[key], self.times[key], data[key], labels[key],
                    sample_weights[key] if sample_weights else None, self.external_feature[key],
                    predict_probability, n_jobs
                )
            return multiple_paths_predictions

        return self._single_backtest_predictions(
            estimator, self.times, data, labels, sample_weights, self.external_feature, predict_probability, n_jobs
        )
